{                         FPKPascal Runtime-Library                         }
{                             DOS Unit for Linux                            }
{ You can set it here, but it should be set through the makefile }
{ $define nativelinux}

Unit Dos;

{$E-}

Interface

Uses Strings
{$ifdef nativelinux}
,syscalls
{$endif}
;

Const
  {Bitmasks for CPU Flags}
  fcarry =     $0001;
  fparity =    $0004;
  fauxiliary = $0010;
  fzero =      $0040;
  fsign =      $0080;
  foverflow  = $0800;

  {Bitmasks for file attribute}
  readonly =   $01;
  hidden =     $02;
  sysfile =    $04;
  volumeid =   $08;
  directory =  $10;
  archive =    $20;
  anyfile =    $3F;
  fmclosed =   $D7B0;
  fminput =    $D7B1;
  fmoutput =   $D7B2;
  fminout =    $D7B3;


Type
  ComStr  = String[255];  { size increased to be more compatible with Unix}
  PathStr = String[255];  { size increased to be more compatible with Unix}
  DirStr  = String[255];  { size increased to be more compatible with Unix}
  NameStr = String[255];  { size increased to be more compatible with Unix}
  ExtStr  = String[255];  { size increased to be more compatible with Unix}


{$PACKRECORDS 1}
  SearchRec = Record
    {Fill : array[1..21] of byte;  Fill replaced with below}

    SearchNum: LongInt; {to track which search this is}
    SearchPos: LongInt; {directory position}
    DirPtr: LongInt; {directory pointer for reading directory}
    SearchType: Byte;  {0=normal, 1=open will close}
    SearchAttr: Byte; {attribute we are searching for}
    Fill: Array[1..07] of Byte; {future use}
    {End of replacement for fill}

    Attr : Byte; {attribute of found file}
    Time : LongInt; {last modify date of found file}
    Size : LongInt; {file size of found file}
    Reserved : Word; {future use}
    Name : String[255]; {name of found file}
    SearchSpec: String[255]; {search pattern}
    NamePos: Word; {end of path, start of name position}
    End;


{$PACKRECORDS 2}
  FileRec = Record
    Handle : word;
    Mode : word;
    RecSize : word;
    _private : array[1..26] of byte;
    UserData: array[1..16] of byte;
    Name: array[0..255] of char;
    End;


  TextBuf = array[0..127] of char;


  TextRec = record
    handle : word;
    mode : word;
    bufSize : word;
    _private : word;
    bufpos : word;
    bufend : word;
    bufptr : ^textbuf;
    openfunc : pointer;
    inoutfunc : pointer;
    flushfunc : pointer;
    closefunc : pointer;
    userdata : array[1..16] of byte;
    name : array[0..255] of char;
    buffer : textbuf;
    End;


    Registers = record
      case i : integer of
        0 : (ax,f1,bx,f2,cx,f3,dx,f4,bp,f5,si,f51,di,f6,ds,f7,es,f8,flags,fs,gs : word);
        1 : (al,ah,f9,f10,bl,bh,f11,f12,cl,ch,f13,f14,dl,dh : byte);
        2 : (eax,  ebx,  ecx,  edx,  ebp,  esi,  edi : longint);
        End;


  DateTime = record
    Year: Word;
    Month: Word;
    Day: Word;
    Hour: Word;
    Min: Word;
    Sec: word;
    End;

{$PACKRECORDS 1}
GTRec = Record
   Year: Word;
   Month: Word;
   MDay: Word;
   WDay: Word;
   Hour: Word;
   Minute: Word;
   Second: Word;
   End;

  Var
    DosError : integer;
    drivestr : array [0..26] of pchar;

Procedure GetDate(var year, month, mday, wday: word);
Procedure GetTime(var hour, minute, second, sec100: word);
Function  DosVersion: Word;
procedure SetDate(year,month,day: word);
Procedure SetTime(hour,minute,second,sec100: word);
Procedure GetCBreak(var breakvalue: boolean);
Procedure SetCBreak(breakvalue: boolean);
Procedure GetVerify(var verify: boolean);
Procedure SetVerify(verify: boolean);
Function  DiskFree(drive: byte) : longint;
Function  DiskSize(drive: byte) : longint;
Procedure FindFirst(const path: pathstr; attr: word; var f: searchRec);
Procedure FindNext(var f: searchRec);
Procedure FindClose(Var f: SearchRec);
Procedure SwapVectors;
Procedure MSDos(var regs: registers);
Procedure GetIntVec(intno: byte; var vector: pointer);
Procedure SetIntVec(intno: byte; vector: pointer);
Procedure Keep(exitcode: word);
Procedure Intr(intno: byte; var regs: registers);
Procedure GetFAttr(var f; var attr: word);
Procedure SetFAttr(var f; attr: word);
Procedure GetFTime(var f; var time: longint);
Procedure SetFTime(var f; time: longint);
Procedure UnpackTime(p: longint; var t: datetime);
Procedure PackTime(var t: datetime; var p: longint);
Function  FSearch(path: pathstr; dirlist: string): pathstr;
Function  FExpand(const path: pathstr): pathstr;
Procedure FSplit(path: pathstr; var dir: dirstr; var name: namestr;
  var ext: extstr);
Procedure Exec(const path: pathstr; const comline: comstr);
Function  DosExitCode: word;
Function  EnvCount: longint;
Function  EnvStr(index: integer): string;
Function  GetEnv (envvar: string): string;
Procedure UnixToDt(SecsPast: LongInt; Var Dt: DateTime);
Function  DTToUnixDate(DT: DateTime): LongInt;
Procedure JulianToGregorian(JulianDN: LongInt; Var Year, Month, Day: Integer);
Function  GregorianToJulian(DT: DateTime): LongInt;
function weekday(y,m,d : longint) : longint;

Implementation

Const RtlFindSize = 15;

Const
   C1970 = 2440588;
   D0 =    1461;
   D1 =  146097;
   D2 = 1721119;

Type  RtlFindRecType = Record
  SearchNum: LongInt;
  DirPtr: LongInt;
  LastUsed: LongInt;
  End;

Type RtlInfoType = Record
   FMode: LongInt;
   FInode: LongInt;
   FUid: LongInt;
   FGid: LongInt;
   FSize: LongInt;
   FMTime: LongInt;
   End;

Var
  RtlFindRecs: Array[1..RtlFindSize] of RtlFindRecType;
  CurrSearchNum: LongInt;
{$ifdef nativelinux}
  envcnt : longint; { Number of environment strings }
  p : ppchar;      { needed to calculate envcnt in initialisation }
{$endif}

{$ifndef nativelinux}
{ Links to C library }
Procedure _rtl_getenv(target: pchar; st: pchar); [ C ];
Procedure _rtl_envstr(i: longint; st: pchar); [ C ];
Function  _rtl_envcnt: longint; [ C ];
Procedure _rtl_gettime(gt: longint); [ C ]; 
Procedure _rtl_getversion(rel: pchar); [ C ];
Function  _rtl_exec(cmdline: pchar; var exitst: integer): integer; [ C ];
Procedure _rtl_closedir(dirptr: longint); [ C ];
Procedure _rtl_seekdir(dirptr: longint; seekpos: longint); [ C ];
Function  _rtl_telldir(dirptr: longint): longint; [ C ];
Function  _rtl_opendir(path: pchar): longint; [ C ];
Procedure _rtl_readdir(dirptr: longint; dname: pchar); [ C ];
Procedure _rtl_stat(path: pchar; infoptr: longint); [ C ];
Procedure _rtl_fstat(fd: longint; infoptr: longint); [ C ];
{$else}

{ The Native replacements for the previous, as far as they're not defined in 
  syscalls.pp }

Function do_exec ( Commandline : pchar; tmp : integer) : integer;

var i,nr : longint;
    p,q : ppchar; { Contains the command line arguments }
    Buf : pchar;
    status : integer;

begin
  { Construct command line args }
  { Count them first }
  nr:=0;
  i:=0; 
  buf:=commandline;
  while (buf^<>#0) do
    begin
    while ((buf^=' ') or (buf^=#8)) or (buf^=#10) do buf:=buf+1;
    {we're on a non-whitespace}
    inc(nr);
    while not (    ((buf^=' ') or (buf^=#8)) 
                or ((buf^=#10) or (buf^=#0))
              ) do buf:=buf+1;
    { We're back on whitespace, or at the end }
    end;
  { nr contains now the nr of args, including the program name }
  getmem(p,nr*4); {get room for the argv}
  if p=nil then
     begin
     do_exec:=0;
     tmp:=8; {Short of Memory}
     exit
     end; 
  i:=0;
  buf:=commandline;
  q:=p; {save p}
  while buf^<>#0 do
    begin
    while ((buf^=' ') or (buf^=#8)) or (buf^=#10) do 
       begin
       buf^:=#0;
       buf:=buf+1
       end;
    {We're on a non-whitespace, hence on an argument}
    p^:=buf;
    p:=p+4; 
    while not (   ((buf^=' ') or (buf^=#8)) 
               or ((buf^=#10) or (buf^=#0))
              ) do buf:=buf+1;
    {We're back on whitespace, or at the end}
    end;
    p^:=nil;
  { q is now argv }
  i:=sys_fork;
  if i=0 then
     begin 
     { The child does the actual exec, and then exits }
     { we use envp, defined in syslinux }
     i:=sys_execve (pchar(q^),ppchar(q),ppchar(envp));
     { If the execve fails, we return an exitvalue of 127, to let it be known }
     halt (127)
     end
  else
    if i=-1 then
       begin
       { Fork failed }
       do_exec:=0;
       tmp:=8;
       exit 
       end;
  { We're in the parent, let's wait.}
  sys_waitpid (i,@status,0);
  if status=127 then
     begin
     { The child couldn't execve !! }
     do_exec:=0;
     tmp:=8 { We set this error, erroneously, since we cannot get to the real error }
     end
  else
     begin
     do_exec:=status shr 8;
     tmp:=0
     end
end;
{$endif}

Procedure Intr (intno: byte; var regs: registers);
  Begin
  { Does not apply to Linux - not implemented }
  End;


Var
  LastDosExitCode: word;

Procedure Exec (Const Path: PathStr; Const ComLine: ComStr);
  Var
    Buff: Array[0..512] of Char;
    Tmp: Integer;

  Begin
  Tmp := Length(Path);
  Move(Path[1], Buff[0], Tmp);
  Buff[Tmp] := ' ';
  Move(ComLine[1], Buff[Tmp + 1], Length(ComLine));
  Buff[Tmp + Length(ComLine) + 1] := #0;
  Tmp := 0;
  {$ifndef nativelinux}
  LastDosExitCode := _rtl_exec(pchar(@buff), Tmp);
  {$else}
  LastDosExitCode := do_exec (pchar(@buff), tmp);
  {$endif}
  Doserror  := Tmp;
  End;


Function DosExitCode: Word;
  Begin
  DosExitCode := LastDosExitCode;
  End;


Function DosVersion: Word;
  Var
    Buffer: Array[0..255] of Char;
    TmpStr: String;
    Tmp2: String;
    Rel: LongInt;
    TmpPos: Word;
    SubRel: LongInt;
{$ifdef nativelinux}
    info : utsname;
{$endif}

  Begin
  {$ifdef nativelinux}
  sys_uname (info);
  Move (info.release,buffer[0],40);
  {$else}
  _rtl_getversion(buffer);
  {$endif}
  Move(Buffer[0], TmpStr[1], 40);
  TmpStr[40] := #0;
  TmpStr[0] := Chr(Pos(#0, TmpStr) -1);
  TmpPos := Pos('.', TmpStr);
  If TmpPos > 0 Then
    Begin
    Tmp2 := Copy(TmpStr, TmpPos + 1, 40);
    If Pos('.', Tmp2) > 0 Then
      Tmp2 := Copy(Tmp2, 1, Pos('.', TmpStr) - 1);
    Val(Tmp2, SubRel); 
    TmpStr := Copy(TmpStr, 1, TmpPos - 1);
    End;
  Val(TmpStr, Rel);
  DosVersion := Rel + (256 * SubRel);
  End;



Procedure GetDate(Var Year, Month, MDay, WDay: Word);
  Var
{$ifndef nativelinux}
    GT: GTRec;
{$else}
    tmp: longint;
    dt : datetime;

{$endif} 
  Begin
{$ifdef nativelinux}
  tmp:=sys_time;
  UnixToDT (tmp,dt);
  year:=dt.year;
  month:=dt.month;
  Mday:=dt.day;
  wday:=weekday (year,month,mday);
{$else}
  _rtl_gettime(longint(@gt));
  Year := gt.year + 1900;
  Month := gt.month + 1;
  MDay := gt.mday;
  WDay := gt.wday;
{$endif}
  End;


Procedure SetDate(Year, Month, Day: Word);
  Begin
  { !! }
  End;


Procedure GetTime(Var Hour, Minute, Second, Sec100: Word);
  Var
{$ifndef nativelinux}
    GT: GTRec;
{$else}
    tmp: longint;
    dt : datetime;

{$endif} 
  Begin
{$ifdef nativelinux}
  tmp:=sys_time;
  UnixToDT (tmp,dt);
  hour:=dt.hour;
  minute:=dt.min;
  second:=dt.sec;
  sec100:=0;
{$else}
  _rtl_gettime(longint(@GT));
  Hour := GT.Hour;
  Minute := GT.Minute;
  Second := GT.Second;
  Sec100 := 0;
{$endif}
  End;


Procedure SetTime(Hour, Minute, Second, Sec100: Word);
  Begin
  { !! }
  End;


Procedure GetCBreak(Var BreakValue: Boolean);
  Begin
  { Not implemented for Linux, but set to true as a precaution. }
  breakvalue:=true
  End;


Procedure SetCBreak(BreakValue: Boolean);
  Begin
  { ! No Linux equivalent ! }
  End;


Procedure GetVerify(Var Verify: Boolean);
   Begin
   { Not implemented for Linux, but set to true as a precaution. }
   verify:=true;
   End;


Procedure SetVerify(Verify: Boolean);
  Begin
  { ! No Linux equivalent ! }
  End;


{
  The diskfree and disksize functions need a file on the specified drive, since this
  is required for the statfs system call.
  These filenames are set in drivestr[0..26], and have been preset to : 
   drivestr[0]^:='.'       (default drive - hence current dir is ok.)
   drivestr[1]^:='/fd0/.'  (floppy drive 1 - should be adapted to local system )
   drivestr[2]^:='/fd1/.'  (floppy drive 2 - should be adapted to local system )
   drivestr[3]^:='/'       (C: equivalent of dos is the root partition)
   drivestr[4..26] are set to nil, but can be set. (you need a getmem (..) !!)
   These settings can be coded in dos.pp, in the initialisation part,
   but can also be set at run time.
   
   They both return -1 when a failure occurs.  
}   

Function DiskFree(Drive: Byte): Longint;

{$ifdef nativelinux}
var fs : statfs;
{$endif}

  Begin
{$ifdef nativelinux}
  if drivestr[Drive]<>nil then 
     begin
     if sys_statfs (pchar(drivestr[drive]),fs)<>0 then 
        begin
        diskfree:=-1; 
        end
     else
        begin
        diskfree:=fs.bavail*fs.bsize; { We could also use bfree ?}
        end
     end
  else
     begin
     diskfree:=-1;
     end
{$endif}
  End;


Function DiskSize(Drive: Byte): Longint;

{$ifdef nativelinux}
var fs : statfs;
{$endif}

  Begin
{$ifdef nativelinux}
  if drivestr[Drive]<>nil then 
     if sys_statfs (pchar(drivestr[drive]),fs)<>0 then 
        disksize:=-1
     else
        disksize:=fs.blocks*fs.bsize
  else
     disksize:=-1;
{$endif}
  End;


Procedure FindClose(Var f: SearchRec);
{ Closes dirptr if it is open }
  Var
    ArrayPos: Word;
    i: Word;
    
  Begin
  ArrayPos := 0;
  If f.SearchType = 0 Then
    Begin
    For i := 1 to RtlFindSize Do
      Begin
      If RtlFindRecs[i].SearchNum = f.SearchNum Then
        ArrayPos := i;
      End;  
    If ArrayPos > 0 Then
      Begin
      RtlFindRecs[ArrayPos].SearchNum := 0;
      {$ifdef nativelinux}
      closedir (pdir(f.dirptr));
      {$else}
      _rtl_closeDir(f.dirptr);
      {$endif}
      End;
    End
  Else
    {$ifdef nativelinux}
    closedir (pdir(f.dirptr));
    {$else}
    _rtl_closedir(f.dirptr);
    {$endif}
  f.dirptr := 0;  
  End;


Function FNMatch(Var Pattern: PathStr; Var Name: PathStr): Boolean;
  Var
    ln: Byte;
    lp: Byte;

  Function DoFNMatch(i: Byte; j: Byte): Boolean;
    Var
      Temp: Boolean;

    Begin
    Temp := True;
    While (i <= lp) and Temp Do
      Begin
      Case pattern[i] of
        '?' : If j > ln then 
                temp:=false;
        '*' :
              Begin
              { find the next character in pattern, different of ? and * }
              inc(i);
              while  (i <= lp) and
              ((pattern[i] =' ?') or (pattern[i] = '*')) and temp do
                begin
                {if it is a ? and there is no more letter in name, it doesn't match}
                if (pattern[i]='?') and (j>ln) then 
                  temp:=false;
                { only increase j if it was a ? }
                if pattern[i]='?' then 
                  inc (j);
                inc (i);
                end;
              { Now, find in name the character which i points to, if the
                * or ? wasn't the last character in the pattern,
                else, use up all the chars in name }
              if i<=lp then
                begin
                repeat
                  begin
                  { find a letter (not only first !) which maches pattern[i] }
                  while (j<=ln) and (name[j]<>pattern[i]) do 
                    inc (j);
                  { if found, compare the rest }
                  if j<=ln then
                    begin
                    temp:=dofnmatch (i+1,j+1);
                    if temp then
                      begin
                      j:=ln+1; { We've found one, we can stop}
                      i:=lp+1
                      end
                    else
                      inc (j) {We didn't find one, need to look further}
                    end
                  else
                    { We didn't find a letter which matches, we can stop }
                    temp:=false;
                  end;
                until (j>ln) or temp;
                end
              else
                j:=ln+1
            end
          else { not a wildcard character in pattern }
            if (j>ln) or (pattern[i]<>name[j]) then 
              temp:=false
        end; { case }
      inc(i);
      inc(j);
      end; {while}
    { at this point, we've exhausted all characters in pattern, or
      we know the thing doesn't match }
    {if we still have unused chars in name, it doesn't match}
    if j<=length(name) then
      dofnmatch:=false
    else
      dofnmatch:=temp
    end;

  Begin {start FNMatch}
  lp := Length(Pattern);
  ln := Length(Name);
  FNMatch := DoFNMatch(1,1);
  End;



Procedure FindWorkProc(Var f: SearchRec);
{ Actual work routine for findfirst/findnext }
  Var
    Found: Boolean;
    Finished: Boolean;
    FName: String[255];
    SName: String[255];
    FBuf: Array[0..255] of Char;
    Info: RtlInfoType;
{$ifdef nativelinux}
    st : stat;
    p : pdirent;
{$endif}
    DT: DateTime;

  Begin
  SName := Copy(f.SearchSpec, f.NamePos + 1, Length(f.SearchSpec)
    - f.NamePos);
  Found := False;
  If f.dirptr = 0 Then
    Finished := True
  Else
    Finished := False;
  While Not Finished Do
    Begin
    {$ifdef nativelinux}
    p:=sys_readdir (pdir(f.dirptr));
    if p=nil then
       fname[1]:=#0
    else
       move (p^.name[0],Fname[1],255);
    {$else}
    _rtl_readdir(f.dirptr, @FBuf);
    Move(FBuf[0], FName[1], 255);
    {$endif}
    FName[255] := #0;
    FName[0] := #255;
    FName[0] := Chr(Pos(#0, FName) - 1);
    If FName = '' Then
      Finished := True
    Else
      Begin
      If FNMatch(SName, FName) Then { !! and attr match}
        Begin
        Move(f.SearchSpec[1], FBuf[0], f.NamePos);
        Move(FName[1], FBuf[f.NamePos], Length(FName));
        FBuf[f.NamePos + Length(FName)] := #0;
        {$ifdef nativelinux}
        sys_stat (@fbuf,st);
        if s_isdir (st.mode) then info.fmode:=16 else info.fmode:=0;
        if ((st.mode and S_IWUSR) = 0) then info.fmode:=(info.fmode or 1);
        info.FSize:=st.size;
        info.FMTime:=st.mtime;
        {$else}
        _rtl_stat(@FBuf, LongInt(@Info));
        {$endif}
        If ((Info.FMode and Not(f.searchattr)) = 0) Then
          Begin
          Finished := True;
          Found := True;
          End;
        End;
      End;
    End; 
  If Found Then
    Begin
    {$ifdef nativelinux}
    f.searchpos:= telldir (pdir(f.dirptr));
    {$else}
    f.searchpos := _rtl_telldir(f.dirptr);
    {$endif}
    DosError := 0;
    f.Name := FName;
    f.Attr := Info.FMode; 
    f.Size := Info.FSize;
    UnixToDT(Info.FMTime, DT);
    PackTime(DT, f.Time);
    End;
  If Not Found Then
    Begin
    FindClose(f);
    DosError := 18;
    End;
  End;


Function  FindLastUsed: Word;
{ Find unused or least recently used dirpointer slot in findrecs array }
  Var
    i: Word;
    Found: Boolean;
    BestMatch: Word;

  Begin
  i := 1;
  Found := False;
  BestMatch := 1;
  While (i < RtlFindSize) And (Not Found) Do
    Begin
    If (RtlFindRecs[i].SearchNum = 0) Then
      Begin
      BestMatch := i;
      Found := True;
      End
    Else
      Begin
      If RtlFindRecs[i].LastUsed > RtlFindRecs[BestMatch].LastUsed Then
        BestMatch := i;
      End;
    Inc(i);
    End;
  FindLastUsed := BestMatch;
  End;


Procedure FindFirst(Const Path: PathStr; Attr: Word; Var f: SearchRec);
{ opens dir and calls FindWorkProc }
  Var
    DirName: Array[0..255] of Char;
    ArrayPos: Word;

  Begin
  f.SearchSpec := Path;
  f.NamePos := 0;
  For ArrayPos := 1 To Length(f.SearchSpec) Do
    Begin
    If f.SearchSpec[ArrayPos] = '/' Then
      f.NamePos := ArrayPos;
    End;
  f.SearchType := 0;
  If f.NamePos > 0 Then
    Begin
    Move(f.SearchSpec[1], DirName[0], f.NamePos);
    DirName[f.NamePos] := #0;
    End
  Else
    Begin
    DirName[0] := '.';
    DirName[1] := '/';
    DirName[2] := #0;
    End;
  {$ifdef nativelinux}
  f.DirPtr := longint(opendir(@(DirName)));
  {$else}
  f.DirPtr := _rtl_opendir(DirName);
  {$endif}
  Inc(CurrSearchNum);
  f.SearchNum := CurrSearchNum;
  f.SearchAttr := Attr;
  If f.DirPtr > 0 Then
    Begin
    For ArrayPos := 1 to RtlFindSize Do
      Inc(RtlFindRecs[ArrayPos].LastUsed);
    ArrayPos := FindLastUsed;
    RtlFindRecs[ArrayPos].LastUsed := 0;
    If RtlFindRecs[ArrayPos].SearchNum > 0 Then
      {$ifdef nativelinux}
      CloseDir(pdir(rtlfindrecs[arraypos].dirptr));
      {$else}
      _rtl_closeDir(rtlfindrecs[arraypos].dirptr);
      {$endif}
    RtlFindRecs[ArrayPos].SearchNum := f.SearchNum;
    RtlFindRecs[ArrayPos].DirPtr := f.DirPtr;
    FindWorkProc(f);
    End
  Else
    DosError := 3;
  End;


Procedure FindNext(Var f: SearchRec);
{ re-opens dir if not already in array and calls FindWorkProc }
Var
  DirName: Array[0..256] of Char;
  ArrayPos: Word;
  i: Word;

  Begin
  If f.SearchType = 0 Then
    Begin
    ArrayPos := 0;
    For i := 1 to RtlFindSize Do
      Begin
      If RtlFindRecs[i].SearchNum = f.SearchNum Then
        ArrayPos := i;
      Inc(RtlFindRecs[i].LastUsed);
      End;
    If ArrayPos = 0 Then
      Begin
      If f.NamePos = 0 Then
        Begin
        DirName[1] := '.';
        DirName[2] := '/';
        DirName[3] := #0;
        End
      Else
        Begin
        Move(f.SearchSpec[1], DirName[0], f.NamePos);
        DirName[f.NamePos] := #0;
        End;
      {$ifdef nativelinux}
      f.DirPtr := longint(opendir(@(DirName)));
      {$else}
      f.DirPtr := _rtl_opendir(DirName);
      {$endif}
      If f.DirPtr > 0 Then
        Begin
        ArrayPos := FindLastUsed;
        End;
      If RtlFindRecs[ArrayPos].SearchNum > 0 Then
         {$ifdef nativelinux}
         CloseDir(pdir(rtlfindrecs[arraypos].dirptr));
         {$else}
         _rtl_closeDir(rtlfindrecs[arraypos].dirptr);
         {$endif}
      RtlFindRecs[ArrayPos].SearchNum := f.SearchNum;
      RtlFindRecs[ArrayPos].DirPtr := f.DirPtr;
      {$ifdef nativelinux}
      seekdir(pdir(f.dirptr), f.searchpos);
      {$else}
      _rtl_seekdir(f.dirptr, f.searchpos);
      {$endif}
      End;
    RtlFindRecs[ArrayPos].LastUsed := 0;
    End;
  FindWorkProc(f);
  End;



Procedure SwapVectors;
  Begin
  { Does not apply to Linux - Do Nothing }
  End;


Function EnvCount: Longint;

  Begin
{$ifdef nativelinux}
  EnvCount := envcnt
{$else}
  EnvCount := _rtl_envcnt;
{$endif}
  End;


Function EnvStr(Index: Integer): String;

  Var
{$ifndef nativelinux}
    Buffer: Array[0..255] of Char;
    TmpStr: String[255];
{$else}
    i : integer;
    p : ppchar;
{$endif}

  Begin
{$ifdef nativelinux}
  p:=envp; { defined in syslinux }
  i:=1;
  while (i<Index) and (p^<>nil) do
    begin
    inc(i);
    p:=p+4;
    end;
  if p=nil then
     envstr:=''
  else
     envstr:=strpas (p^)
{$else}
  Buffer[0] := #0;
  _rtl_envstr(index, buffer);
  Move (Buffer, TmpStr[1], 255);
  TmpStr[0] := #255;
  TmpStr[255] := #0;
  TmpStr[0] := Chr(Pos(#0, TmpStr) - 1);
  EnvStr := TmpStr;
{$endif}
  End;


Function GetEnv(EnvVar: String): String;

var 
{$ifdef nativelinux}
    ep : ppchar;
    gotit : boolean;
    l : byte;
{$else}
    Buffer: Array[0..255] of Char;
    OutStr: Array[0..255] of Char;
    TmpStr: String[255];
    Tmp: LongInt;
{$endif}

  Begin
{$ifdef nativelinux}
  ep:=envp;
  {Bug fix, before, the first variable starting with 'envvar' would be found.
   now, the correct one is found.}
  envvar:=envvar+'=';
  l:=length(envvar);
  gotit:=false;
  while (ep^<>nil) and not gotit do 
     if strlcomp(@envvar[1],(ep^),l)=0 then
        gotit:=true
     else
        ep:=ep+4;
  if gotit then
     getenv:=strpas (ep^+l)
  else
     getenv:=''
{$else}
  OutStr[0] := #0;
  Move(EnvVar[1], Buffer, Length(EnvVar));
  Buffer[Length(EnvVar)] := #0;
  _rtl_getenv(buffer, outstr);
  Move(OutStr, TmpStr[1], 255);
  TmpStr[0] := #255;
  TmpStr[255] := #0;
  TmpStr[0] := Chr(Pos(#0, TmpStr) - 1);
  GetEnv := TmpStr;
  {
  Tmp := Pos('=', TmpStr);
  If Tmp > 0 Then
    GetEnv := Copy(TmpStr, Tmp + 1, 255)
  Else
    GetEnv := '';
  }
{$endif}
  End;

Procedure FSplit(Path: PathStr; Var Dir: DirStr; Var Name: NameStr;
  Var Ext: ExtStr);

  Var
    DotPos: Integer;
    SlashPos: Integer;
    i: Integer;
    
  Begin
  SlashPos := 0;
  DotPos := 0;
  i := Length(Path);
  While i > 0 Do
    Begin
    If (Path[i] = '.') And (DotPos = 0) And (SlashPos = 0) Then
      DotPos := i;
    If (Path[i] = '/') And (SlashPos = 0) Then
      SlashPos := i;
    Dec(i);
    End;
  If DotPos > SlashPos Then
    Begin
    Ext := Copy(Path, DotPos, 255);
    If SlashPos > 0 Then
      Begin
      Dir := Copy(Path, 1, SlashPos);
      Name := Copy(Path, SlashPos + 1, DotPos - SlashPos - 1);
      End
    Else
      Begin
      Dir := '';
      Name := Copy(Path, 1, DotPos - 1);
      End;
    End 
  Else
    Begin
    Ext := '';
    If SlashPos > 0 Then
      Begin
      Dir := Copy(Path, 1, SlashPos);
      Name := Copy(Path, SlashPos + 1, Length(Path) - SlashPos - 1);
      End
    Else
      Begin
      Dir := '';
      Name := Path;
      End;
    End;
  End;


Function FExpand(Const Path: PathStr): PathStr;

var temp : pathstr;
    i,j    : byte;

  Begin
  { Remove eventual drive - doesn't exist in Linux }
  if path[2]=':' then i:=3 else i:=1;
  { Do we have an absolute path ? No - prefix the current dir }
  if path[i]<>'/' then 
        getdir (0,temp)
  else
    begin
    inc (i);
    temp:=''
    end;
  temp:=temp+'/'+copy(path,i,length(path)-i+1);
  { First remove all references to '/./'}
  while pos ('/./',temp)<>0 do
     delete (temp,pos('/./',temp),2);
  { Now remove also all references to '/../' + of course previous dirs.. }
  i:=pos('/../',temp);
  while i<>0 do 
    begin
    { Find the pos of the previous dir }
    if i>1 then
       begin
       j:=i-1;
       while (j>1) and (temp[j]<>'/') do dec (j); { temp[1] is always '/' }
       delete (temp,j,i-j+3)
       end 
    else
       { i=1, so we have temp='/../something', just delete '/../' }
       delete (temp,1,3);
    { Compute new i }
    i:=pos('/../',temp);
    end;   
   fexpand:=temp;
  End;

Procedure msdos(var regs : registers);
  Begin
  { ! Not implemented in Linux ! }
  End;


Procedure getintvec(intno : byte;var vector : pointer);
  Begin
  { ! Not implemented in Linux ! }
  End;


Procedure setintvec(intno : byte;vector : pointer);
  Begin
  { ! Not implemented in Linux ! }
  End;


Procedure keep(exitcode : word);
  Begin
  { ! Not implemented in Linux ! }
  End;


Procedure getfattr(var f; var attr : word);
  Var
{$ifdef nativelinux}
    info : stat;
{$else}
    Info: RtlInfoType;
{$endif}

  Begin
{$ifdef nativelinux}
  if sys_stat (pchar (f),info)<>0 then 
      begin
      attr:=0;
      doserror:=3
      end
  else 
      attr:=info.mode; 
{$else}  
  _rtl_fstat(longint(f), longint(@Info));
  attr := info.fmode;
{$endif}
  End;


Procedure setfattr (var f;attr : word);
  Begin
  { ! Not implemented in Linux ! }
  End;


Procedure getftime (var f; var time : longint);

  Var
{$ifndef nativelinux}
    Info: RtlInfoType;
{$else}
    len : word;
    info : stat;
{$endif}
    DT: DateTime;

  Begin
{$ifndef nativelinux}
  _rtl_fstat(longint(f), longint(@Info));
  UnixToDT(Info.FMTime, DT);
{$else}
  len:=ord(filerec(f).name[0]);
  filerec(f).name[len+1]:=#0;
  if sys_stat (pchar (@filerec(f).name[1]),info)<>0 then 
     begin 
     Time:=0;
     doserror:=3;
     exit
     end
  else 
     UnixToDT(Info.mTime, DT);
{$endif}
  PackTime(DT, Time);
  End;


Procedure setftime(var f; time : longint);
  Begin
  { ! Not implemented in Linux ! }
  End;


Procedure unpacktime(p : longint;var t : datetime);
  Begin
  t.sec:= (p mod 32)*2;
  p:=p div 32;
  t.min:= p mod 64;
  P:=p div 64;
  t.hour := p mod 32;
  p := p div 32;
  t.day:=p mod 32;
  p:=p div 32;
  T.month:= p mod 16;
  p:=p div 16;
  t.year := p +1980;
  End;


Procedure packtime(var t : datetime;var p : longint);
  Begin
  p := (t.day + (32 * t.month) + (512 * (t.year - 1980))) * 65536;
  P := p + (t.sec div 2) + (32 * t.min) + (2048 * t.hour);
  End;


Function  fsearch(path : pathstr;dirlist : string) : pathstr;
  Var NewDir: PathStr;
    p1: Byte;
    s: SearchRec;

  Begin
  If (Pos('?', Path) <> 0) or (Pos('*', Path) <> 0) Then
    FSearch := '' {No wildcards allowed in these things.}
  Else
    Begin
    Dirlist:='.;'+dirlist; { Make sure current dir is first to be searched. }
    Repeat
      p1 := Pos(';', DirList);
      If p1 > 0 Then
        Begin
        NewDir := Copy(DirList, 1, P1 - 1);
        Delete(DirList, 1, p1)
        End
      Else
        Begin
        NewDir := DirList;
        DirList := ''
        End;
      FindFirst(NewDir + '/' + Path, AnyFile, s);
      If DosError = 0 Then
        Begin
        NewDir := NewDir + '/' +s.name;
        If Pos('./', NewDir) = 1 Then 
          Delete(NewDir, 1, 2)
        {DOS strips off an initial .\}
        End
      Else
        NewDir := '';
    Until(DirList = '') or (Length(NewDir) > 0);
    FSearch := NewDir;
    End
  End;


Function GregorianToJulian(DT: DateTime): LongInt;
Var
  Century: LongInt;
  XYear: LongInt;
  Temp: LongInt;
  Month: LongInt;

  Begin
  Month := DT.Month;
  If Month <= 2 Then
    Begin
    Dec(DT.Year);
    Inc(Month,12);
    End;
  Dec(Month,3);
  Century := DT.Year Div 100;
  XYear := DT.Year Mod 100;
  Century := (Century * D1) shr 2;
  XYear := (XYear * D0) shr 2;
  GregorianToJulian :=  ((((Month * 153) + 2) div 5) + DT.Day) + D2
    + XYear + Century;
  End;


Procedure JulianToGregorian(JulianDN: LongInt; Var Year, Month, Day: Integer);
  Var
    Temp,
    XYear: LongInt;
    YYear,
    YMonth,
    YDay: Integer;

  Begin
  Temp := (((JulianDN - D2) shl 2) - 1);
  XYear := (Temp Mod D1) or 3;
  JulianDN := Temp Div D1;
  YYear := (XYear Div D0);
  Temp := ((((XYear mod D0) + 4) shr 2) * 5) - 3;
  YMonth := Temp Div 153;
  If YMonth >= 10 Then
    Begin
    YYear := YYear + 1;
    YMonth := YMonth - 12;
    End;
  YMonth := YMonth + 3;
  YDay := Temp Mod 153;
  YDay := (YDay + 5) Div 5;
  Year := YYear + (JulianDN * 100);
  Month := YMonth;
  Day := YDay;
  End;


Procedure UnixToDt(SecsPast: LongInt; Var Dt: DateTime);
  Var
    DateNum: LongInt;

  Begin
  Datenum := (SecsPast Div 86400) + c1970;
  JulianToGregorian(DateNum, Integer(DT.Year), Integer(DT.Month),
    Integer(DT.day));
  SecsPast := SecsPast Mod 86400;
  DT.Hour := SecsPast Div 3600;
  SecsPast := SecsPast Mod 3600;
  DT.Min := SecsPast Div 60;
  DT.Sec := SecsPast Mod 60;
  End;


Function DTToUnixDate(DT: DateTime): LongInt;
   Var
     SecsPast, DaysPast: LongInt;

  Begin
  DaysPast := GregorianToJulian(DT) - c1970;
  SecsPast := DaysPast * 86400;
  SecsPast := SecsPast + (LongInt(DT.Hour) * 3600) + (DT.Min * 60) + (DT.Sec);
  DTToUnixDate := SecsPast;
  End;
  
function weekday(y,m,d : longint) : longint;

{ Calculates th day of the week. Florian provided this.
  returns -1 on error }

    var
       century_offset : byte;   
       temp : longint;
       _is_leap_year : boolean; 
       
    const
       month_table : array[1..12] of longint = (1,4,4,0,2,5,0,3,6,1,4,6);
       
  function is_leap_year(y : longint) : boolean;
  
    begin
       if (y mod 100)=0 then
         is_leap_year:=((y mod 400)=0)
       else
         is_leap_year:=(y mod 4)=0;
    end;

  { Beginning of weekday }  
  begin
       if (m<1) or (m>12) then
         begin
            weekday:=-1;
            exit;
         end;
       case y of
          1700..1799 : century_offset:=4;
          1800..1899 : century_offset:=2;
          1900..1999 : century_offset:=0;
          2000..2099 : century_offset:=-1;
          else
            begin
               if (y>=2100) then
                 begin
                 end;
               weekday:=-1;
               exit;
            end;  
       end;
       _is_leap_year:=is_leap_year(y);
       y:=y mod 100;
       temp:=(y div 12)+(y mod 12)+((y mod 12) div 4);
       temp:=temp mod 7;
       temp:=(temp+month_table[m]+d) mod 7;
       { do some corrections for special years }
       { other century ? }
       inc(temp,century_offset);
       { leap year correction }
       if _is_leap_year and (m<3) then
         dec(temp);
       { now is sonday 1, but should be for example 0 }
       dec(temp);
       { the result could be less than zero }
       while temp<0 do
         inc(temp,7);
       weekday:=temp mod 7;
    end;
    
Begin
For CurrSearchNum := 1 To RtlFindSize Do
  Begin
  RtlFindRecs[CurrSearchNum].SearchNum := 0;
  RtlFindRecs[CurrSearchNum].LastUsed := 0;
  End;
CurrSearchNum := 0;
{$ifdef nativelinux}
{ Calculate the number of environment strings}
envcnt:=0;
p:=envp;
while (p^<>nil) do
  begin
  inc (envcnt);
  p:=p+4
  end;
{ drivestr needs initialisation }
getmem (drivestr[0],3);
strpcopy (drivestr[0],'.');            { Default drive }
getmem (drivestr[1],8);
strpcopy (drivestr[1],'/fd0/.');       { Floppy drve 1 (a:) }
getmem (drivestr[2],8);
strpcopy (drivestr[2],'/fd1/.');       { floppy drive 2 (b:) }
getmem (drivestr[3],3);
strpcopy (drivestr[3],'/');            { Hard disk }
{$endif} 
End.