#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/utsname.h>
#include <sys/wait.h>
#include <dirent.h>

/* #define RTLDEBUG */
/* #define AOUT */

void PASCALMAIN();

typedef struct
   {
   short int gt_year;
   short int gt_month;
   short int gt_day;
   short int gt_dow;
   short int gt_hour;
   short int gt_min;
   short int gt_sec;
   }
   gt_struct;

typedef struct
   {
   int   fmode;
   int   finode;
   int   fuid;
   int   fgid;
   int   fsize;
   int   fmtime;
   }
   fileinfo_struct;

int crtl_argc;
char **crtl_argv;
extern char **environ; 

#ifdef AOUT
int rtl_errno()
#else
int _rtl_errno()
#endif
/* return error number */
  {
  return(errno);
  }
  
#ifdef AOUT
int rtl_time()
#else
int _rtl_time()
#endif
/* returns seconds since 1970 */
  {
  return(time(0));
  }

#ifdef AOUT
void rtl_exit(long l)
#else
void _rtl_exit(long l)
#endif
/* exit program */
  {
  #ifdef RTLDEBUG
  puts("called rtl_exit");
  #endif
  exit(l);
  }


#ifdef AOUT
long rtl_paramcount()
#else
long _rtl_paramcount()
#endif

/* count of params on command line */
  {
  #ifdef RTLDEBUG
  puts("rtl_paramcount called");
  #endif
  return(crtl_argc);
  }

#ifdef AOUT
void rtl_paramstr(char *st, long l)
#else
void _rtl_paramstr(char *st, long l)
#endif
/* return specific param from command line */
  {
  #ifdef RTLDEBUG
  puts("rtl_paramstr called");
  #endif
  if ( l < 0 || l > crtl_argc)
    {
    *st = (char)0;
    }
  else
    {
    strncpy(st, crtl_argv[l], 255);
    }
  }


#ifdef AOUT
long rtl_open(char *f, long flags)
#else  
long _rtl_open(char *f, long flags)
#endif
/* open a file */
  {
  int oflags;
  
  #ifdef RTLDEBUG
  puts("rtl_open called");
  puts("trying to open");
  puts(f);
  #endif
  if (flags & 0x100) 
    {
    oflags = O_CREAT | O_TRUNC;
    }
  else
    {
    oflags = 0;
    }
  switch (flags & 3)
    {
    case 00:
      {
      oflags = oflags | O_RDONLY;
      break;
      }
    case 01:
      {   
      oflags = oflags | O_WRONLY;
      break;
      }
    case 02:
      {
      oflags = oflags | O_RDWR;
      break;
      }  
    }  
  return(open(f, oflags, 438)); /* mode 666 */
  
  /* open returns a file descriptor or -1 on error 
     flags available as params are
         O_RDONLY 	read only
         O_WRONLY	write only
         O_RDWR		read/write
         O_APPEND	set file offset to eof prior to each write
         O_CREAT	create if it currently doesn't exist
         O_EXCL		on create fail if file exists
         O_NOCTTY	prevent file from becoming control tty
         O_NONBLOCK	don't wait on not ready
         O_TRUNC	truncate file to zero length
         
    returns errors in errno
         EACCES permission problem
         EEXIST file already exists
         EINTR interrupted by signal
         EISDIR attempt to open dir for write or rename file to dir
         EMFILE too many file descriptors
         ENAMETOOLONG name too long
         ENFILE too many files open
         ENOENT does not exist
         ENOSPC no space on disk
         ENOTDIR not a directory
         ENXIO no such device or offline
         EROFS read only file system
  */         
  }

  
#ifdef AOUT  
void rtl_close(long h)
#else  
void _rtl_close(long h)
#endif
/* close a file */
  {
  #ifdef RTLDEBUG
  puts("rtl_close called");
  #endif
  close(h);
  }


#ifdef AOUT
void rtl_write(long h, long addr, long len)
#else
void _rtl_write(long h, long addr, long len)
#endif
/* write to a file */
  {
  #ifdef RTLDEBUG
  puts("rtl_write called");
  #endif
  write(h, (char *)addr, len);
  }


#ifdef AOUT
void rtl_erase(char *p)
#else
void _rtl_erase(char *p)
#endif
/* erase/unlink a file */
  {
  #ifdef RTLDEBUG
  puts("rtl_erase called");
  #endif
  unlink(p);
  }
  

#ifdef AOUT  
void rtl_rename(char *p1, char *p2)
#else
void _rtl_rename(char *p1, char *p2)
#endif
/* rename/move a file */
  {
  #ifdef RTLDEBUG
  puts("rtl_rename called");
  #endif
  rename(p1, p2);
  }
  

#ifdef AOUT  
long rtl_read(int h, char *addr, int len)
#else
long _rtl_read(int h, char *addr, int len)
#endif
/* read from a file */
  {
  #ifdef RTLDEBUG
  puts("rtl_read called");
  #endif
  return(read(h, addr, len));
  }
  
#ifdef AOUT  
long rtl_filepos(long handle)
#else
long _rtl_filepos(long handle)
#endif
/* get current file position */
  {
  #ifdef RTLDEBUG
  puts("rtl_filepos called");
  #endif
  return(lseek(handle, 0 , 1));
  }
  
  
#ifdef AOUT  
void rtl_seek(long handle,long pos)
#else
void _rtl_seek(long handle,long pos)
#endif
/* seek to specific position in file */
  {
  #ifdef RTLDEBUG
  puts("rtl_seek called");
  #endif
  lseek(handle, pos, 0);
  }
  

#ifdef AOUT  
long rtl_filesize(long handle)
#else
long _rtl_filesize(long handle)
#endif
/* return size of file */
  {
  long tmp;
  long fsize;
  
  #ifdef RTLDEBUG
  puts("rtl_filesize called");
  #endif
  tmp = lseek(handle, 0, 1);
  fsize = lseek(handle, 0, 2);
  lseek(handle, tmp, 0);
  return(fsize);
  }
  
  
#ifdef AOUT
void rtl_rmdir(char *buffer)
#else
void _rtl_rmdir(char *buffer)
#endif
/* remove directory */
  {
  #ifdef RTLDEBUG
  puts("rtl_rmdir called");
  #endif
  rmdir(buffer);
  }


#ifdef AOUT  
void rtl_mkdir(char *buffer)
#else
void _rtl_mkdir(char *buffer)
#endif
/* make directory */
  {
  #ifdef RTLDEBUG
  puts("rtl_mkdir called");
  #endif
  mkdir(buffer, 511);  /* mode 777 */
  }
  

#ifdef AOUT
void rtl_chdir(char *buffer)
#else
void _rtl_chdir(char *buffer)
#endif
/* change current working directory */
  {
  #ifdef RTLDEBUG
  puts("rtl_chdir called");
  #endif
  chdir(buffer);
  }
  
  
#ifdef AOUT
void rtl_getenv(char *target, char *st)
#else
void _rtl_getenv(char *target, char *st)
#endif
  {
  char *tmp;
  
  #ifdef RTLDEBUG
  puts("rtl_getenv called");
  #endif
  tmp = getenv(target);
  if (tmp == 0)
    {
    *st = 0;
    }
  else
    {
    strncpy(st, getenv(target), 255);
    }
  }
  
  
#ifdef AOUT
void rtl_envstr(long i, char *st)
#else
void _rtl_envstr(long i, char *st)
#endif
  {
  #ifdef RTLDEBUG
  puts("rtl_envstr called");
  #endif
  strncpy(st, environ[i - 1], 255);
  }  
  

#ifdef AOUT
long rtl_envcnt()
#else
long _rtl_envcnt()
#endif
  {
  long i;
  
  #ifdef RTLDEBUG
  puts("rtl_envcnt called");
  #endif
  i = 0;
  while (environ[i] != NULL)  
    i++;
  return(i + 1);  
  }  


#ifdef AOUT
void rtl_gettime(gt_struct *gt)
#else
void _rtl_gettime(gt_struct *gt)
#endif
  {
  struct tm *t;
  time_t now;

  #ifdef RTLDEBUG
  printf("rtl-gettime called\n");
  #endif
  now = time(0);
  t = localtime(&now);
  gt->gt_year = t->tm_year; 
  gt->gt_month = t->tm_mon;
  gt->gt_day = t->tm_mday;
  gt->gt_dow = t->tm_wday;
  gt->gt_hour = t->tm_hour;
  gt->gt_min = t->tm_min;
  gt->gt_sec = t->tm_sec;
  }
  

#ifdef AOUT
void rtl_getversion(char *rel)
#else
void _rtl_getversion(char *rel)
#endif
  {
  struct utsname *nm;

  nm = malloc(sizeof(struct utsname));
  uname(nm);
  strncpy(rel, nm->release, 255);
  free(nm);
  }


#ifdef AOUT
short int rtl_exec(char *cmdline, short int *exitcode)
#else
short int _rtl_exec(char *cmdline, short int *exitcode)
#endif
  {
  int execcode;
  char *args[64];
  char *buf;
  int pid;
  int exitstatus;
  int numargs;
  
  execcode = 0;
  exitcode = 0;
  numargs = 0;
  buf = cmdline;
  
  /* parse cmd line arguments */
  while (*buf != '\0')
    {
    while ((*buf == ' ') || (*buf == '\t') || (*buf == '\n'))
      *buf++ = '\0';
    args[numargs] = buf;
    numargs ++;
    while ((*buf != '\0') && (*buf != ' ') && (*buf != '\n'))
      buf++;
    }
  args[numargs] = '\0';
  
  /* fork */
  if ((pid = fork()) < 0)
    {
    execcode = 8;
    exit(1);
    }

  /* child process execs cmdline */
  if (pid == 0)
    {
    execvp(*args, args);
    switch (errno)
      {
      case E2BIG:
      case ENOMEM:
      case ENOEXEC:
        {
        execcode = 8;
        break;
        }
      case EACCES:
        {
        execcode = 5;
        break;
        }
      case ENAMETOOLONG:
      case ENOENT:
        {
        execcode = 2;
        break;
        }
      case ENOTDIR:
        {
        execcode = 3;
        break;
        }
      default:
        {
        execcode = 5;
        break;
        }  
      }   
    exit(1);
    }
    
  /* parent process waits for child to finish */
  while (wait(&exitstatus) != pid);
  execcode = WEXITSTATUS(exitstatus);
  /* return exit status */
  return(execcode); /* exit code */
  }


#ifdef AOUT
void rtl_closedir(DIR *dirptr)
#else
void _rtl_closedir(DIR *dirptr)
#endif
  {
  closedir(dirptr);
  }


#ifdef AOUT
void rtl_seekdir(DIR *dirptr, int spos)
#else
void _rtl_seekdir(DIR *dirptr, int spos)
#endif
  {
  seekdir(dirptr, spos);
  }
  

#ifdef AOUT
int rtl_telldir(DIR *dirptr)
#else
int _rtl_telldir(DIR *dirptr)
#endif
  {
  return(telldir(dirptr));
  }

  
#ifdef AOUT
void rtl_readdir(DIR *dirptr, char *fname)
#else
void _rtl_readdir(DIR *dirptr, char *fname)
#endif
  {
  struct dirent *oneent;
  oneent = readdir(dirptr);
  if (oneent == 0)
    *fname = 0;
  else
    strncpy(fname, oneent->d_name, 255);
  }


#ifdef AOUT
DIR *rtl_opendir(char *path)
#else
DIR *_rtl_opendir(char *path)
#endif
  {
  return(opendir(path));
/* open dir returns pointer or null if an error occurs
   errno can be set to
     EACCESS permission problem
     ENAMETOOLONG name too long
     ENOENT does not exits
     ENOTDIR not a directory
     EMFILE too many file descriptors open
     ENFILE too many files open
*/    
  }  

#ifdef AOUT
void rtl_stat(char *path, fileinfo_struct *finfo)
#else
void _rtl_stat(char *path, fileinfo_struct *finfo)
#endif
  {
  struct stat stinfo;
  struct tm   *tmptime;
  time_t      tmp1;
  
  stat(path, &stinfo);
  finfo->fmode = stinfo.st_mode;
  finfo->finode = stinfo.st_ino;
  finfo->fuid = stinfo.st_uid;
  finfo->fgid = stinfo.st_gid;
  finfo->fsize = stinfo.st_size;
  tmptime = localtime(&stinfo.st_mtime);
  tmp1 = mktime(tmptime);
  if (tmptime->tm_isdst > 0)
    tmp1 = tmp1 + 3600;        /* Daylight Savings Time Adj */
  tmptime = gmtime(&stinfo.st_mtime);
  tmp1 = tmp1 - mktime(tmptime);
  finfo->fmtime = stinfo.st_mtime + tmp1;
  finfo->fmode = 0;
  if (S_ISDIR(stinfo.st_mode))
    finfo->fmode = finfo->fmode + 16;
  if ((S_IWUSR & stinfo.st_mode) == 0)
    finfo->fmode = finfo->fmode + 1;
  }
  
  
#ifdef AOUT
void rtl_fstat(int filedesc, fileinfo_struct *finfo)
#else
void _rtl_fstat(int filedesc, fileinfo_struct *finfo)
#endif
  {
  struct stat stinfo;
  struct tm   *tmptime;
  time_t      tmp1;
  
  fstat(filedesc, &stinfo);
  finfo->fmode = stinfo.st_mode;
  finfo->finode = stinfo.st_ino;
  finfo->fuid = stinfo.st_uid;
  finfo->fgid = stinfo.st_gid;
  finfo->fsize = stinfo.st_size;
  tmptime = localtime(&stinfo.st_mtime);
  tmp1 = mktime(tmptime);
  if (tmptime->tm_isdst > 0)
    tmp1 = tmp1 + 3600;        /* Daylight Savings Time Adj */
  tmptime = gmtime(&stinfo.st_mtime);
  tmp1 = tmp1 - mktime(tmptime);
  finfo->fmtime = stinfo.st_mtime + tmp1;
  finfo->fmode = 0;
  if (S_ISDIR(stinfo.st_mode))
    finfo->fmode = finfo->fmode + 16;
  if ((S_IWUSR & stinfo.st_mode) == 0)
    finfo->fmode = finfo->fmode + 1;
  }
  
  
int main(int argc, char *argv[])
/* main c routine calls main pascal routine after saving cmd line info */
  {
  crtl_argc = argc - 1;
  crtl_argv = argv;
  #ifdef AOUT
  asm ("call PASCALMAIN");
  #else
  PASCALMAIN();
  #endif
  return(0);
  }
  
