{$ifndef SYS_LINUX}
unit syscalls;

{
  Unit syscalls; provides an interface to Linux System Calls
  Michael Van Canneyt, 1996

  The ifndef SYS_LINUX statements are used to exclude 
  all unnecessary stuff from compiling when syscalls 
  is included in syslinux.pp

  If you add a new function, include it in ifndef SYS_LINUX
}

{ Set this if you want debugging messages }
{ $define syscall_debug}


interface

uses strings;
{$endif}

{$i sysnr.inc} { All system call numbers for 1.2.13 }

Const 
  { Things for LSEEK call }
  Seek_set = 0;
  Seek_Cur = 1;
  Seek_End = 2;
  { Things for OPEN call - after linux/fcntl.h }
  Open_Accmode  = 3;
  Open_RdOnly   = 0;
  Open_WrOnly   = 1;
  Open_RdWr     = 2;
  Open_Creat    = 1 shl 6;
  Open_Excl     = 2 shl 6;
  Open_NoCtty   = 4 shl 6;
  Open_Trunc    = 1 shl 9;
  Open_Append   = 2 shl 9;
  Open_NonBlock = 4 shl 9;
  Open_NDelay   = Open_NonBlock;
  Open_Sync     = 1 shl 12;
  { The waitpid uses the following options : }
  Wait_NoHang   = 1;
  Wait_UnTraced = 2;
  Wait_Any      = -1;
  Wait_MyPGRP   = 0;
  { Constants to check stat.mode }
  S_IFMT   = $f000; {00170000 }
  S_IFSOCK = $c000;  {0140000}
  S_IFLNK  = $a000;  {0120000}
  S_IFREG  = $8000;  {0100000}
  S_IFBLK  = $6000;  {0060000}
  S_IFDIR  = $4000;  {0040000}
  S_IFCHR  = $2000;  {0020000}
  S_IFIFO  = $1000;  {0010000}
  S_ISUID  = $0800;  {0004000}
  S_ISGID  = $0400;  {0002000}
  S_ISVTX  = $0200;  {0001000}
  { Constants to check permissions }
  S_IRWXO = $7;
  S_IROTH = $4;
  S_IWOTH = $2;
  S_IXOTH = $1;
  
  S_IRWXG = S_IRWXO shl 3;
  S_IRGRP = S_IROTH shl 3;
  S_IWGRP = S_IWOTH shl 3;
  S_IXGRP = S_IXOTH shl 3;
  
  S_IRWXU = S_IRWXO shl 6;
  S_IRUSR = S_IROTH shl 6;
  S_IWUSR = S_IWOTH shl 6;
  S_IXUSR = S_IXOTH shl 6;

  { Constants to test the type of filesystem }
  fs_old_ext2 = $ef51;
  fs_ext2     = $ef53;
  fs_ext      = $137d;
  fs_iso      = $9660;
  fs_minix    = $137f;
  fs_minix_30 = $138f;
  fs_minux_V2 = $2468;
  fs_msdos    = $4d44;
  fs_nfs      = $6969;
  fs_proc     = $9fa0;
  fs_xia      = $012FD16D; 
 
{ Inlude the error-definitions}
{$i errno.inc}

type
{ Linux system calls take arguments as follows :
   %eax : System call number
   %ebx : first argument
   %ecx : second argument
   %edx : third argumens
   %esi : fourth argument
   %edi : fifth argument
  That is why we define a special type, with only these arguments }

 syscallregs = record
      eax,ebx,ecx,edx,esi,edi : longint;
     end;

{ The following are records for system calls }
{$packrecords 1}

  dirent = record
    ino    : longint;
    off    : longint;
    reclen : word;
    name   : string[255]
    end;
  pdirent = ^dirent;

  dir = record
    fd : integer;
    loc : longint; 
    size : integer;
    buf : pdirent;
    {The following are used in libc, but NOT in the linux kernel sources ??}
    nextoff : longint;
    dd_max : integer; {size of buf. Irrelevant, as buf is of type dirent}       
    lock : pointer;
    end;
  pdir = ^dir;

  stat = record
     dev    : word;
     pad1   : word;
     ino    : longint;
     mode   : word;
     nlink  : word;
     uid    : word;
     gid    : word;
     rdev   : word;
     pad2   : word;
     size   : longint;
     blksze : Longint;
     blocks : Longint;
     atime  : Longint;
     unused1 : longint;
     mtime   : Longint;
     unused2 : longint;
     ctime   : Longint;
     unused3 : longint;
     unused4 : longint;
     unused5 : longint;
     end;

   statfs = record
     fstype   : longint; { File system type }
     bsize    : longint; { Optimal block trensfer size }
     blocks   : longint; { Data blocks in system }
     bfree    : longint; { free blocks in system }
     bavail   : longint; { Available free blocks to non-root users }
     files    : longint; { File nodes in system }
     ffree    : longint; { Free file nodes in system }
     fsid     : longint; { File system ID }
     namelen  : longint; { Maximum name length in system }
     spare    : array [0..6] of longint; { For later use }
     end; 
 
  sysinfo = record 
        uptime : longint;                   
        loads : array[1..3] of longint;      
        totalram :  longint;
        freeram : longint;
        sharedram : longint;
        bufferram : longint;
        totalswap : longint;
        freeswap : longint;
        procs : integer;
        s : string[18];
        end;

  time = record
    sec,usec : longint
    end;
  ptime=^time;
 
  timezone = record
    minuteswest,dsttime : integer
    end;
  ptimezone =^timezone;
 
  utsname = record
     sysname,nodename,release,version,machine,domainname  : string[64];
  end;  

  Tpipe = array[1..2] of longint;


{$packrecords normal}

var errno : longint; { We report errors in it }

{ List of system Calls As implemented }

{$ifndef SYS_LINUX}
{ We don't need the declarations when we're in syslinux.pp}

{ Skeleton function, is available if you want to do it yourself }
Function syscall ( callnr : longint; var regs : syscallregs ) : longint;

{ Not really system calls, but related to it }
{ Can be found in utilities section }
Function CloseDir ( p : pdir) : integer;
Function OpenDir (f : pchar) : pdir;
Procedure SeekDir (p : pdir;off : longint);
Function TellDir ( p : pdir) : longint;
{ Check stat.mode for type of file : }
Function S_ISLNK (m : integer) : boolean;
Function S_ISREG(m : integer) : boolean;
Function S_ISDIR(m : integer) : boolean;	
Function S_ISCHR(m : integer) : boolean;
Function S_ISBLK(m : integer) : boolean;
Function S_ISFIFO(m : integer): boolean;
Function S_ISSOCK (m : integer)	: boolean;

{ Implemented System Calls }

Function Sys_Chdir (Filename : pchar) : longint;
Function Sys_Close (f : longint) : longint;
Function Sys_Execve (Filename : pchar; argv : ppchar; envp : ppchar) : Longint;
{ The following is not a function, since it shouldn't return !! }
Procedure Sys_Exit  (ExitCode : Integer);
Function Sys_Fork : longint; 
Function Sys_Fstat (Fd : integer; var Buffer: stat) : longint;
Function Sys_Fstatfs (Fd : integer; var Buffer: statfs) : longint;
Function Sys_GeteGid : Longint;
Function Sys_GeteUid : Longint;
Function Sys_GetGid : longint;
Function Sys_GetPid : Longint;
Function Sys_GetPpid : Longint;
Function Sys_GetTimeOfDay (var timeval : time; var timez : timezone) : integer; 
Function Sys_GetUid : Longint;
Function Sys_Link (oldname : pchar; newname: pchar) : longint;
Function Sys_Lseek (F : Longint; Off : longint; Whence : longint) : Longint;
Function Sys_Lstat (Filename : pchar; var Buffer: stat) : longint;
Function Sys_Mkdir (Filename : pchar; mode : Longint) : longint;
Function Sys_Open(f: pchar; flags: Longint; mode : integer): Longint;
Function Sys_Pipe ( var P : tpipe ) : integer;
Function Sys_Read (f : longint; buffer : pchar; count : longint) : longint;
Function Sys_ReadDir ( p : pdir) : pdirent;
Function Sys_Rename (Oldname : pchar; Newname : pchar) : longint;
Function Sys_Rmdir (Filename : pchar) : longint;
Function Sys_Stat (Filename : pchar; var Buffer: stat) : Longint;
Function Sys_statfs ( Filename : pchar; var buffer : statfs ) : longint;
Function Sys_Symlink (oldname : pchar; newname: pchar) : longint;
Function Sys_Sysinfo (var Inforec : Sysinfo) : Longint;
Function Sys_Time : Longint;
Function Sys_Uname (var Unamerec : Utsname) : Longint;
Function Sys_Unlink (Filename : pchar) : longint;
Function Sys_WaitPid (pid : longint; status : pointer; options : integer) : longint;
Function Sys_Write (f : longint; buffer : pchar; count : longint) : longint;

Implementation
{$endif}

{ 
All calls are sorted according to function :
  - General   : generall calls
  - File      : File handling related calls
  - Directory : Directory related calls
  - Process   : Process & user-info related calls
  - Utilities : Not really system calls, but related
To get to a section quickly, do a search '--- Sectionname' 
}
 

{ --- General : General calls --- }
    
Procedure Do_SysCall ( callnr : longint; var regs : syscallregs );

{ This function puts the registers in place, does the call, and then
 copies back the registers as they are after the syscall.}

begin
asm
# load the registers...
  movl 12(%ebp),%eax
  movl 4(%eax),%ebx
  movl 8(%eax),%ecx
  movl 12(%eax),%edx
  movl 16(%eax),%esi
  movl 20(%eax),%edi
# set the call number
  movl 8(%ebp),%eax
# Go !
  int $0x80
# Put back the registers...
  pushl %eax
  movl 12(%ebp),%eax
  movl %edi,20(%eax)
  movl %esi,16(%eax)
  movl %edx,12(%eax)
  movl %ecx,8(%eax)
  movl %ebx,4(%eax)
  popl %ebx
  movl %ebx,(%eax)
end; { end of assembler }
end;

Function SysCall ( callnr : longint; var regs : syscallregs ) : longint;

{ This function serves as an interface to do_syscall.
  If the syscall returned a negative number, it returns -1, and puts the
  syscall result in errno. Otherwise, it returns the syscall return value }

begin
  do_syscall (callnr,regs);
  if regs.eax<0 then
     begin
     errno:=-regs.eax;
     syscall:=-1
     end
  else
     begin
     syscall:=regs.eax;
     errno:=0
     end
end;

Function Sys_Time : Longint;

var t     : syscallregs;
    dummy : longint;

begin
  t.ebx:=0;
  dummy:=syscall (syscall_nr_time,t);
{$ifdef syscall_debug}
  writeln ('Time syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_Time:=dummy;
end;

{$ifndef SYS_LINUX}
Function Sys_GetTimeOfDay (var timeval : time; var timez : timezone) : integer; 

var t     : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(@timeval);
  t.ecx:=longint(@timez);
  dummy:=syscall (syscall_nr_gettimeofday,t);
{$ifdef syscall_debug}
  writeln ('GetTimeOfDay syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_GetTimeOfDay:=dummy;
end;

Function Sys_Sysinfo (var Inforec : Sysinfo) : Longint;

var t : syscallregs;
    dummy : longint;

Begin
  t.ebx :=longint(@(inforec));
  dummy := syscall(syscall_nr_sysinfo,t);
{$ifdef syscall_debug}
  writeln ('Sysinfo Syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Sysinfo:=dummy;
End;

Function Sys_Uname (var unamerec : utsname) : Longint;

var t : syscallregs;
    dummy : longint;

Begin
  t.ebx :=longint(@(unamerec));
  dummy := syscall(syscall_nr_uname,t);
{$ifdef syscall_debug}
  writeln ('Uname syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Uname:=dummy;
End;
{$endif}


{ --- File : File handling related calls --- }

Function Sys_Open(f: pchar; flags: Longint; mode : integer): Longint;

var t : syscallregs;
    dummy : longint;

Begin
  t.ebx :=longint(f);
  t.ecx :=flags;
  t.edx :=mode;            
  dummy := syscall(syscall_nr_open,t);
{$ifdef syscall_debug}
  writeln ('Open Syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Open:=dummy;
End;

Function Sys_Lseek (F : Longint; Off : longint; Whence : longint) : Longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=f;
  t.ecx:=off;
  t.edx:=Whence;
  dummy:=syscall (syscall_nr_lseek,t);
{$ifdef syscall_debug}
  writeln ('Read syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_lseek:=dummy;
end;


Function Sys_Read (f : longint; buffer : pchar; count : longint) : longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=f;
  t.ecx:=longint(buffer);
  t.edx:=count;
  dummy:=syscall (syscall_nr_read,t);
{$ifdef syscall_debug}
  writeln ('Read syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Read:=dummy;
end;


Function Sys_Write (f : longint; buffer : pchar; count : longint) : longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=f;
  t.ecx:=longint(buffer);
  t.edx:=count;
  dummy:=syscall (syscall_nr_write,t);
{$ifdef syscall_debug}
  writeln ('Write syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Write  := dummy;
end;

Function Sys_Close (f : longint) : longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=f;
  dummy:=syscall (syscall_nr_close,t);
{$ifdef syscall_debug}
  writeln ('Close syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Close:=dummy;
end;

Function Sys_Unlink (Filename : pchar) : longint;

var t     : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(filename);
  dummy:=syscall (syscall_nr_unlink,t);
{$ifdef syscall_debug}
  writeln ('Unlink syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Unlink :=dummy;
end;

Function Sys_Rename (Oldname : pchar; Newname : pchar) : longint;

var t     : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(oldname);
  t.ecx:=longint(newname);
  dummy:=syscall (syscall_nr_rename,t);
{$ifdef syscall_debug}
  writeln ('Rename syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Rename :=dummy;
end;

{ We need this for getcwd }
Function Sys_Stat (Filename : pchar; var Buffer: stat) : Longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(filename);
  t.ecx:=longint(@(buffer));
  dummy:=SysCall (syscall_nr_stat,t);
{$ifdef syscall_debug}
  writeln ('Stat syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_Stat:=dummy;
end;

{$ifndef SYS_LINUX}
Function Sys_Fstat (Fd : integer; var Buffer: stat) : longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=Fd;
  t.ecx:=longint(@(buffer));
  dummy:=SysCall (syscall_nr_fstat,t);
{$ifdef syscall_debug}
  writeln ('Fstat syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_fstat:=dummy;
end;

Function Sys_Fstatfs (Fd : integer; var Buffer: statfs) : longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=Fd;
  t.ecx:=longint(@(buffer));
  dummy:=SysCall (syscall_nr_fstatfs,t);
{$ifdef syscall_debug}
  writeln ('Fstatfs syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_fstatfs:=dummy;
end;

Function Sys_Lstat (Filename : pchar; var Buffer: stat) : longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(filename);
  t.ecx:=longint(@(buffer));
  dummy:=SysCall (syscall_nr_lstat,t);
{$ifdef syscall_debug}
  writeln ('Lstat syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_Lstat:=dummy;
end;

Function Sys_statfs (Filename : pchar; var Buffer: statfs) : longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(filename);
  t.ecx:=longint(@(buffer));
  dummy:=SysCall (syscall_nr_statfs,t);
{$ifdef syscall_debug}
  writeln ('Statfs syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_statfs:=dummy;
end;


Function Sys_Link (oldname : pchar; newname: pchar) : longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(oldname);
  t.ecx:=longint(newname);
  dummy:=SysCall (syscall_nr_link,t);
{$ifdef syscall_debug}
  writeln ('Link syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_Link:=dummy;
end;

Function Sys_Pipe ( var P : tpipe ) : integer;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(@p);
  dummy:=syscall (syscall_nr_pipe,t);
{$ifdef syscall_debug}
  writeln ('Pipe syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  sys_pipe:=dummy
end;
{$endif}

{ We need this for erase ... }

Function Sys_Symlink (oldname : pchar; newname: pchar) : longint;

var t : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(oldname);
  t.ecx:=longint(newname);
  dummy:=SysCall (syscall_nr_symlink,t);
{$ifdef syscall_debug}
  writeln ('SymLink syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_symlink:=dummy;
end;


{ --- Directory : Directory related calls --- } 

Function Sys_Mkdir (Filename : pchar; mode : Longint) : longint;

var t     : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(filename);
  t.ecx:=mode;
  dummy:=syscall (syscall_nr_mkdir,t);
{$ifdef syscall_debug}
  writeln ('Mkdir syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Mkdir :=dummy;
end;

{ We need this for getcwd } 
Function Sys_ReadDir ( p : pdir) : pdirent;

var t     : syscallregs;
    dummy : longint;
 
begin
  t.ecx:=longint(p^.buf);
  t.ebx:=p^.fd;
  t.edx:=1;
  dummy:=syscall (syscall_nr_readdir,t);
{$ifdef syscall_debug}
  writeln ('Readdir syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  { the readdir system call returns the number of bytes written }
  if dummy=0 then 
     sys_readdir:=nil 
  else 
     sys_readdir:=p^.buf
end;


Function Sys_Rmdir (Filename : pchar) : longint;

var t     : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(filename);
  dummy:=syscall (syscall_nr_rmdir,t);
{$ifdef syscall_debug}
  writeln ('Rmdir  syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Rmdir :=dummy;
end;


Function Sys_Chdir (Filename : pchar) : longint;

var t     : syscallregs;
    dummy : longint;

begin
  t.ebx:=longint(filename);
  dummy:=syscall (syscall_nr_chdir,t);
{$ifdef syscall_debug}
  writeln ('Chdir syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Chdir :=dummy;
end;

{ --- Process : Process & program handling - related calls --- }

{$ifndef SYS_LINUX}
Function Sys_Execve (Filename : pchar; argv : ppchar; envp : ppchar) : Longint;

var t : syscallregs;
    dummy : Longint;

begin
  t.ebx:=longint (Filename);
  t.ecx:=longint (argv);
  t.edx:=longint (envp);
  dummy:=syscall (syscall_nr_execve,t);
{ The rest normally is only executed when an error occurs. }
{$ifdef syscall_debug}
  writeln ('Execve Syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Execve:=dummy;
End;
{$endif}

Procedure Sys_Exit  (ExitCode : Integer);

var t : syscallregs;

begin
  t.ebx:=exitcode;
  syscall (syscall_nr_exit,t)
end;

{$ifndef SYS_LINUX} 
Function Sys_Fork : longint; 

var t : syscallregs;
    dummy : longint;

Begin
  dummy := syscall(syscall_nr_fork,t);
{$ifdef syscall_debug}
  writeln ('Fork Syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_Fork:=dummy;
End;

Function Sys_GetPid : Longint;

var t : syscallregs;
    dummy : longint;

begin
  dummy:=SysCall (syscall_nr_getpid,t);
{$ifdef syscall_debug}
  writeln ('GetPid syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_GetPid:=dummy;
end;

Function Sys_GetPpid : Longint;

var t : syscallregs;
    dummy : longint;

begin
  dummy:=SysCall (syscall_nr_getppid,t);
{$ifdef syscall_debug}
  writeln ('GetPpid syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_GetPpid:=dummy;
end;

Function Sys_GetUid : Longint;

var t : syscallregs;
    dummy : longint;

begin
  dummy:=SysCall (syscall_nr_getuid,t);
{$ifdef syscall_debug}
  writeln ('GetUid syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_GetUid:=dummy;
end;

Function Sys_GeteUid : Longint;

var t : syscallregs;
    dummy : longint;

begin
  dummy:=SysCall (syscall_nr_geteuid,t);
{$ifdef syscall_debug}
  writeln ('GetEuid syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_GetEuid:=dummy;
end;

Function Sys_Getgid : Longint;

var t : syscallregs;
    dummy : longint;

begin
  dummy:=SysCall (syscall_nr_getgid,t);
{$ifdef syscall_debug}
  writeln ('Getgid syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_Getgid:=dummy;
end;

Function Sys_Getegid : Longint;

var t : syscallregs;
    dummy : longint;

begin
  dummy:=SysCall (syscall_nr_getegid,t);
{$ifdef syscall_debug}
  writeln ('GetEgid syscall returned : ',dummy);
  writeln ('Errno = ',errno);
{$endif}
  Sys_GetEgid:=dummy;
end;

Function Sys_WaitPid (pid : longint; status : pointer; options : integer) : longint;

var t : syscallregs;
    dummy : longint;

Begin
  t.ebx := pid;
  t.ecx := longint(status);
  t.edx := options;
  dummy := syscall(syscall_nr_waitpid,t);
{$ifdef syscall_debug}
  writeln ('Waitpid syscall returned : ',dummy);
  writeln ('Errno : ',errno);
{$endif}
  Sys_WaitPid:=dummy;
End;
{$endif}

{ --- Utilities : Not really system calls }

function CloseDir ( p : pdir) : integer;

begin
  closedir:=sys_close (p^.fd)
end;

{ we need this for getcwd }
Function OpenDir (f : pchar) : pdir;

var fd : integer;
    st : stat;
    ptr : pdir;

begin
  if sys_stat (f,st)<0 then
     begin
     opendir:=nil;
     exit
     end;
  { Is it a dir ? }
  if not ((st.mode and $f000) = $4000) then 
    begin
    errno:=sys_enotdir;
    opendir:=nil;
    exit
    end;
  { Open it}
  fd:=sys_open (f,OPEN_RDONLY,438);
  if fd<0 then
    begin
    opendir:=nil;
    exit
    end;
  new(ptr);
  if ptr=nil then
    begin
    opendir:=nil;
    exit
    end;
  new (ptr^.buf);
  if ptr^.buf=nil then
    begin
    opendir:=nil;
    exit
    end;
  ptr^.fd:=fd;
  ptr^.loc:=0;
  ptr^.size:=0;
  ptr^.dd_max:=sizeof (ptr^.buf^);
  opendir:=ptr;
end;

{$ifndef SYS_LINUX}
procedure SeekDir (p : pdir;off : longint);

begin
  if p=nil then
     begin
     errno:=sys_EBADF;
     exit
     end;
  p^.nextoff := sys_lseek (p^.fd,off,seek_set);
  p^.size:=0;
  p^.loc:=0;
end;

function TellDir ( p : pdir) : longint;

begin
  if p=nil then
     begin
     errno:=sys_EBADF;
     telldir:=-1;
     exit
     end;
  telldir:=sys_lseek (p^.fd,0,seek_cur)
  { We could try to use the nextoff field here, but on my 1.2.13 
    kernel, this gives nothing... This may have to do with 
    the readdir implementation of libc... I also didn't find any trace of
    the field in the kernel code itself, So I suspect it is an artifact of libc.
    Michael. }
end;
{$endif}
{$ifndef SYS_LINUX}
Function S_ISLNK (m : integer) : boolean;
begin
 S_ISLNK:=(m and S_IFMT) = S_IFLNK
end;

Function S_ISREG(m : integer) : boolean;
begin
 S_ISREG:=(m and S_IFMT) = S_IFREG
end;

Function S_ISDIR(m : integer) : boolean;	
begin
 S_ISDIR:=(m and S_IFMT) = S_IFDIR
end;

Function S_ISCHR(m : integer) : boolean;
begin
  S_ISCHR:=(m and S_IFMT) = S_IFCHR
end;

Function S_ISBLK(m : integer) : boolean;
begin
  S_ISBLK:=(m and S_IFMT) = S_IFBLK
end;

Function S_ISFIFO(m : integer): boolean;
begin
  S_ISFIFO:=(m and S_IFMT) = S_IFIFO
end;

Function S_ISSOCK (m : integer)	: boolean;
begin
  S_ISSOCK:=(m and S_IFMT) = S_IFSOCK
end;
{$endif}

{$ifndef SYS_LINUX}
begin
  errno:=0
end.
{$endif}
