Unit Crt;

{ -------------------------------------------------------------------
  CRT unit for FPK-Pascal.
  Written by Mark May (mmay@ndaco.com)
  Major overhaul to improve screen output performance by 
  Michael Van Canneyt  (michael@tfdec1.fys.kuleuven.ac.be)
  Also added some documentation.
  -------------------------------------------------------------------}

Interface

Uses dos,syscalls; 

Const
  Black = 0;
  Blue = 1;
  Green = 2;
  Cyan = 3;
  Red = 4;
  Magenta = 5;
  Brown = 6;
  LightGray = 7;
  DarkGray = 8;
  LightBlue = 9;
  LightGreen = 10;
  LightCyan = 11;
  LightRed = 12;
  LightMagenta = 13;
  Yellow = 14;
  White = 15;
  Blink = 128;


Const
  TextAttr: Byte = $07;
  TextChar: Char = ' ';
  CheckBreak: Boolean = True;
  CheckEOF: Boolean = False;
  CheckSnow: Boolean = False;
  DirectVideo: Boolean = False;
  LastMode: Word = 3;
  WindMin: Word = $0;
  WindMax: Word = $184f;
  ScreenWidth = 80;
  ScreenHeight = 25;


Procedure AssignCrt(Var F: Text);
Procedure Delay(DTime: Word);
Procedure TextColor(CL: Byte);
Procedure TextBackground(CL: Byte);
Procedure GoToXy(X: Byte; Y: Byte);
Function  WhereX: Byte;
Function  WhereY: Byte;
Procedure Window(X1, Y1, X2, Y2: Byte);
Procedure ClrScr;
Procedure ClrEol;
Function  KeyPressed: Boolean;
Function  ReadKey: Char;
Function  SaveScreenRegion(xl,yl,xh,yh: Byte; Var Pt: Pointer):Boolean;
Procedure RestoreScreenRegion(xl,yl,xh,yh: Byte; Pt: Pointer);
Procedure ScrollWindow(xl,yl,xh,yh : Byte; count: Longint);
Procedure HighVideo;
Procedure LowVideo;
Procedure NormVideo;
Procedure DelLine;
Procedure InsLine;
Procedure DoIbmPcChars(B: Boolean);
Function  PopKeyChar: Char;
Procedure PushKeyChar(Ch: Char);


Implementation

Const
  KeyBufferSize = 20;

Type TScreenLine = Array[1..ScreenWidth] of char;
     TScreen = Array[1..ScreenHeight] of TSCreenLine;
     TLineColors = Array[1..ScreenWidth] of byte;
     TScreenColors = Array[1..ScreenHeight] of TLineColors;
     
Var
  Scrn : TScreen;
  ScrnCol : TScreenColors;
  
  KeyBuffer: Array[1..KeyBufferSize] of Char;
  KeyPut: Byte;
  KeySend: Byte;
  CrtFOut: Integer;
  CrtFIn: Integer;
  CurrX: Byte;
  CurrY: Byte;
  ExitSave: Pointer;
  EmptyLine: String[ScreenWidth];

{ ------------------------------------------------------------------- }
{  Utility routines }

Procedure DumpString (Str: String);
{ 
  Write a string directly to the screen 
}

Begin
  Sys_Write(LongInt(CrtFOut), @Str[1], Length(Str));
End;

Function GenerateXY(X: Byte; Y: Byte) : String;
{
  Returns a string with the escape sequences to go to X,Y on the screen
}
Var
  StX: String[3];
  StY: String[3];

Begin
  Str(x, StX);
  Str(y, StY);
  GenerateXY := #27 + '[' + StY + ';' + StX + 'f';
End;

Function CvtColor(colr:Byte):String;
{
  Returns a string with the color number to switch the display 
  to the desired color. (DOS colors<> VT100 colors)
}
  Begin
  Colr := Colr mod 8;
  Case Colr of
    0: cvtcolor := '0';
    1: cvtcolor := '4';
    2: cvtcolor := '2';
    3: cvtcolor := '6';
    4: cvtcolor := '1';
    5: cvtcolor := '5';
    6: cvtcolor := '3';
    7: cvtcolor := '7';
    End;
  End;

Function GenerateCol (Attr : Word) : String;
{
  Returns a string with the formatting to set the desired fore and
  background color.  
}

  Var
    DoReset: Boolean;
    DoBlink: Boolean;
    DoHigh: Boolean;
    DoFore: Boolean;
    DoBack: Boolean;
    TmpStr: String;

Begin
  DoReset := true;
  DoBlink := ((Attr and $80) <> 0) And DoReset;
  DoHigh  := ((Attr and $08) <> 0) and DoReset;
  DoFore  := ((Attr and $07) <> 0);
  DoBack  := ((Attr and $70) <> 0);
  TmpStr := #27 + '[';
  If DoReset Then
    TmpStr := TmpStr + '0;';
  If DoBlink Then
    TmpStr := TmpStr + '5;';
  If DoHigh Then
    TmpStr := TmpStr + '1;';
  If DoFore Then
    Begin
      TmpStr := TmpStr + '3' + CvtColor(Attr and $07) + ';'
    End;
  If DoBack Then
    TmpStr := TmpStr + '4' + CvtColor((Attr shr 4) and $07) + ';' ;
  TmpStr[Length(TmpStr)] := 'm';
  GenerateCol := TmpStr;
End;

Function WinMinX: Byte;
{
  Current Minimum X coordinate
}
  Begin
  WinMinX := (WindMin and $ff) + 1;
  End;


Function WinMaxX: Byte;
{
  Current Maximum X coordinate
}
  Begin
  WinMaxX := (WindMax and $ff) + 1;
  End;


Function WinMinY: Byte;
{
  Current Minimum Y Coordinate 
}
  Begin
  WinMinY := (WindMin shr 8) + 1;
  End;


Function WinMaxY: Byte;
{
  Current Maximum Y coordinate;
}
  Begin
  WinMaxY := (WindMax shr 8) + 1;
  End;



Procedure Delay(DTime: Word);
{
  Wait for DTime milliseconds.
}
  Var
    TRec: Time;
    SelectArray: Array[1..5] of LongInt;
    Regs: SysCallRegs;
   
  Begin
  TRec.Sec := 0;
  TRec.USec := DTime * 1000;
  SelectArray[1] := 1;
  SelectArray[2] := 0;
  SelectArray[3] := 0;
  SelectArray[4] := 0;
  SelectArray[5] := LongInt(@TRec);
  Regs.ebx := LongInt(@SelectArray);
  SysCall(Syscall_nr_select, regs);
  End;


Procedure TextColor(CL: Byte);
{
  Switch foregroundcolor
}
  Begin
  TextAttr := TextAttr and $F0;
  TextAttr := TextAttr or (CL and $0F);
  DumpString(GenerateCol(TextAttr));
  End;


Procedure TextBackground(CL: Byte);
{
  Switch backgroundcolor
}
  Begin
  TextAttr := TextAttr and $0F;
  TextAttr := TextAttr or (CL shl 4);
  DumpString(GenerateCol(TextAttr));
  End;

Procedure GoToXy(X: Byte; Y: Byte);
{
  Go to coordinates X,Y in the current window.
}
  Begin
  If (X > 0) and (X <= WinMaxX + 1 - WinMinX) and
  (Y > 0) and (Y <= WinMaxY + 1 - WinMinY) Then
    Begin
    CurrX := X + (WindMin and $ff);
    CurrY := Y + (WindMin shr 8);
    DumpString (GenerateXY(CurrX, CurrY));
    Dec(CurrX);
    End;
  End;


Procedure Window(X1, Y1, X2, Y2: Byte);
{
  Set screen window to the specified coordinates.
}
  Begin
  if (X1>X2) or (X2>ScreenWidth) then exit;
  if (Y1>Y2) or (Y2>ScreenHeight) then exit;
  WindMin := (Y1 - 1);
  WindMin := (WindMin Shl 8) + (X1 - 1);
  WindMax := (Y2 - 1);
  WindMax := (WindMax Shl 8) + (X2 - 1);
  GoToXY(1,1);
  End;


Procedure ClrScr;
{
  Clear the current window, and set the cursor on x1,y1
}
  Var
    CY: Integer;
    TmpStr: String;

  Begin
  For Cy := WinMinY To WinMaxY Do
      Begin
      TmpStr:=GenerateXY(WinMinX,Cy)+copy(EmptyLine,1,WinMaxX-WinMinX+1);
      DumpString(TmpStr);
      { Update memory copy }
      Move (EmptyLine[1],Scrn[Cy,WinMinX],WinMaxX-WinMinX+1);
      FillChar (ScrnCol[Cy,WinMinX],WinMaxX-WinMinX+1,char(TextAttr));
      End;
  GoToXY(1,1);
  End;


Procedure ClrEol;
{
  Clear from current position to end of line.
}
  Var

  Begin
  DumpString(Copy(EmptyLine,1,WinMaxX-CurrX));
  GotoXY (CurrX+1,CurrY)
  End;


Function WhereX: Byte;
{
  Return current X-position of cursor.
}
  Begin
  WhereX := CurrX - (WindMin and $ff)+1;
  End;


Function WhereY: Byte;
{
  Return current Y-position of cursor.
}
  Begin
  WhereY := CurrY - (WindMin shr 8);
  End;

Procedure ScrollScrnRegionUp(xl,yl,xh,yh, count: Byte);
{
  Scroll the indicated region count lines up. The empty lines are filled
  with blanks in the current color. The screen position is restored 
  afterwards.
}
  Var
    Temp : string[255];
    i: Byte;
    j: Byte;
    TA : word;
  Begin
  TA:=TextAttr;
  For i := yl To yh-count Do
    Begin
    Temp:=GenerateXY(xl,i);
    For j := xl To xh Do
      Begin
      if ScrnCol[i+count,j]<>TextAttr then
        begin
        TextAttr:=ScrnCol[i+count,j];
        temp:=temp+GenerateCol(TextAttr);
        end;
      Temp:=Temp+Scrn[i+count,j];
      { Update memory copy }
      Scrn[i,j]:=Scrn[i+count,j];
      ScrnCol[i,j]:=ScrnCol[i+count,j];
      if length(Temp)>240 then
        begin
        DumpString(Temp);
        Temp:='';
        end; 
      End;
    If length(Temp)>0 then DumpString(Temp); 
    End;
  { Restore Current Color}
  TextAttr:=TA;
  Temp:=GenerateCol(TextAttr);
  DumpString(temp);
  { Fill the rest with empty lines. }
  For i:=yh-count+1 to yh do
    begin
    temp:=GenerateXY(xl,i)+copy(EmptyLine,1,xh-xl+1);
    DumpString(temp);
    { Update memory copy }
    Move (EmptyLine[1],Scrn[i,xl],xh-xl+1);
    FillChar (ScrnCol[i,xl],xh-xl+1,TextAttr);
    end;
  { Restore current position }
  GotoXY(CurrX+1,CurrY);
End;


Procedure ScrollScrnRegionDown(xl,yl,xh,yh, count: Byte);
{
  Scroll the indicated region count lines down. The empty lines are filled
  with blanks in the current color. The screen position is restored 
  afterwards.
}

  Var
    TA,i,j : Byte;
    temp   : String;
    
  Begin
  TA:=TextAttr;
  for i:=yh downto yl+count do
    begin
    temp:=GenerateXY(xl,i);
    for j:=xl to xh do
      begin
      if ScrnCol[i-count,j]<>TextAttr then
        begin
        TextAttr:=ScrnCol[i-count,j];
        temp:=temp+GenerateCol(TextAttr);
        end;
      temp:=temp+Scrn[i-count,j];
      { Update memory copy }
      Scrn[i,j]:=Scrn[i-count,j];
      ScrnCol[i,j]:=TextAttr;
      if length(Temp)>240 then 
        begin
        DumpString(Temp);
        Temp:='';
        end;
      end;
    If length(Temp)>0 then DumpString(temp);
    end;
  TextAttr:=TA;
  Dumpstring(GenerateCol(TextAttr));    
  for i:=yl to yl+count-1 do
    begin
    temp:=GenerateXY(xl,i)+Copy(EmptyLine,1,xh-xl+1);
    DumpString(temp);
    { Update memory copy }
    Move (EmptyLine[1],Scrn[i,xl],xh-xl+1);
    FillChar (ScrnCol[i,xl],xh-xl+1,TextAttr);
    end;
  GotoXY(CurrX+1,CurrY);
end;

Procedure ScrollWindow(xl,yl,xh,yh : Byte; count: LongInt);
{
  Scroll the indicated region up or down, depending on the sign
  of count.
}
begin
  If (xl>xh) or (xh>ScreenWidth) then exit;
  If (yl>yh) or (yl>ScreenHeight) then exit;
  If (abs(Count)>yh-yl+1) then exit;
  If count<0 then
    ScrollScrnRegionDown (xl,yl,xh,yh,abs(count))
  else
    ScrollScrnRegionUp (xl,yl,xh,yh,count);
end;

Function SaveScreenRegion(xl,yl,xh,yh: Byte; Var Pt: Pointer):Boolean;
{
  If pointer<>nil then allocate enough space to hold the specified
  region. The region is stored as a sequence of 2 bytes. First byte
  is the character code, second byte is the color code.
  The Region is stored line per line.
  returns false if no memory was available to store the region in.
}
  Var
    Tx: Byte;
    Ty: Byte;
    Ctr: Word;

  Begin
  If Pt = nil then GetMem(Pt, ((xh + 1 - xl) * (yh +1 - yl) * 2));
  If Pt = nil Then
    SaveScreenRegion := False
  Else
    Begin
    SaveScreenRegion := True;
    Ctr := 0;
    For Ty := yl to yh Do
      For Tx := xl to xh Do
        Begin
        Pchar(Pt)[Ctr] := Scrn[Ty, Tx];
        inc(Ctr);
        Pchar(Pt)[Ctr] := char(ScrnCol[Ty,Tx]);
        Inc(Ctr);
        End;;
    End;
  End;


Procedure RestoreScreenRegion(xl,yl,xh,yh: Byte; Pt: Pointer);
{
  Restores a region with data from pt. 
}
  Var
    Tx: Byte;
    Ty: Byte;
    Ctr: Word;

  Begin
{
  If Pt <> nil Then
    Begin
    Ctr := 0;
    For Ty := yl to yh Do
      Begin
      For Tx := xl to xh Do
        Begin
        PutScrnWord(Tx, Ty, WordArrayPtr(PT)^[Ctr]);
        Inc(Ctr);
        End;
      End;
    FreeMem(Pt, ((xh + 1 - xl) * (yh +1 - yl) * 2));
    End;
}
  End;

Function  RealKeyPressed: Boolean;
{
  Low-level function to see if a key was pressed.
}
  Var
    TRec: Time;
    SelectArray: Array[1..5] of LongInt;
    Regs: SysCallRegs;
    FDS: Array[0..31] of LongInt;
    i: LongInt;
   
  Begin
  TRec.Sec := 0;
  TRec.USec := 1000;
  For i := 0 to 31 Do
    FDS[i] := 0;
  FDS[0] := 1;
  SelectArray[1] := 1;
  SelectArray[2] := LongInt(@FDS);
  SelectArray[3] := 0;
  SelectArray[4] := 0;
  SelectArray[5] := LongInt(@TRec);
  Regs.ebx := LongInt(@SelectArray);
  RealKeyPressed := (SysCall(Syscall_nr_select, regs) <> 0);
  End;


Function  KeyPressed: Boolean;
{
  Check in the buffer if a key was pressed. If not, check at the
  kernel level.
}
  Begin
  If KeyPut <> KeySend Then
    KeyPressed := True
  Else
    KeyPressed := RealKeyPressed;
  End;


Function  NumKeysBuffed: Word;
{
  Returns the number of keys stored in the buffer.
}
  Var
    Tmp: Integer;

  Begin
  Tmp := KeyPut - KeySend;
  If Tmp < 0 Then
    Begin
    Inc(Tmp, KeyBufferSize)
    End;
  NumKeysBuffed := Tmp;
  End;


Function  PopKeyChar: Char;
{
  Returns the first key stored in the buffer
}
  Begin
  If KeyPut <> KeySend Then
    Begin
    PopKeyChar := KeyBuffer[KeySend];
    Inc(KeySend);
    If KeySend > KeyBufferSize Then
      KeySend := 1;
    End
  Else
    PopKeyChar := #0;
  End;


Procedure PushKeyChar(Ch: Char);
{
  Pushes an extra key in the keybuffer.
}
  Var
    Tmp: Word;

  Begin
  Tmp := KeyPut;
  Inc(KeyPut);
  If KeyPut > KeyBufferSize Then
    KeyPut := 1;
  If KeyPut <> KeySend Then
    KeyBuffer[Tmp] := Ch
  Else
    KeyPut := Tmp;
  End;


Procedure PushAltKey(TmpChar: Char);
{
  Push an alt key in the buffer.
}
  Begin
    Case TmpChar Of
      'q': PushKeyChar(#016);
      'w': PushKeyChar(#017);
      'e': PushKeyChar(#018);
      'r': PushKeyChar(#019);
      't': PushKeyChar(#020);
      'y': PushKeyChar(#021);
      'u': PushKeyChar(#022);
      'i': PushKeyChar(#023);
      'o': PushKeyChar(#024);
      'p': PushKeyChar(#025);
      'a': PushKeyChar(#030);
      's': PushKeyChar(#031);
      'd': PushKeyChar(#032);
      'f': PushKeyChar(#033);
      'g': PushKeyChar(#034);
      'h': PushKeyChar(#035);
      'j': PushKeyChar(#036);
      'k': PushKeyChar(#037);
      'l': PushKeyChar(#038);
      'z': PushKeyChar(#044);
      'x': PushKeyChar(#045);
      'c': PushKeyChar(#046);
      'v': PushKeyChar(#047);
      'b': PushKeyChar(#048);
      'n': PushKeyChar(#049);
      'm': PushKeyChar(#050);
      '1': PushKeyChar(#120);
      '2': PushKeyChar(#121);
      '3': PushKeyChar(#122);
      '4': PushKeyChar(#123);
      '5': PushKeyChar(#124);
      '6': PushKeyChar(#125);
      '7': PushKeyChar(#126);
      '8': PushKeyChar(#127);
      '9': PushKeyChar(#128);
      '0': PushKeyChar(#129);
      '-': PushKeyChar(#130);
      '=': PushKeyChar(#131);
    End;
  End;



Procedure ReadKeyDtl;
{
 Read keys from the kernel.
}
  Var
    TmpChar: Char;
    NumRead: LongInt;
    State: Word;
    Finished: Boolean;

  Begin
  State := 0;
  Finished := False;
  NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
  Case TmpChar of
    #27: If RealKeyPressed Then
           State := 1;
    #10: TmpChar := #13;
    #127: TmpChar := #08;
  End;
  If State > 0 Then
    Begin
    While (RealKeyPressed and Not Finished) Do
      Begin
      NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
      Case State of
        1: Begin { have esc }
           Case TmpChar of
             'a'..'z','0'..'9','-','=': 
               Begin
               PushKeyChar(#0);
               PushAltKey(TmpChar);
               Finished := True;
               End;
             '[':
               Begin
               State := 2;
               End;
             Else
               Begin
               PushKeyChar(#27);
               PushKeyChar(TmpChar);
               State := 0;
               End;
             End;
           End;
        2: Begin { have esc [ }
           Case TmpChar of
             '[': Begin
               State := 3;
               End;
             'A': Begin
               PushKeyChar(#0);
               PushKeyChar(#72);
               End;
             'B': Begin
               PushKeyChar(#0);
               PushKeyChar(#80);
               End;
             'C': Begin
               PushKeyChar(#0);
               PushKeyChar(#77);
               End;
             'D': Begin
               PushKeyChar(#0);
               PushKeyChar(#75);
               End;
             '5': Begin
               PushKeyChar(#0);
               PushKeyChar(#73);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;               
               End;
             '6': Begin
               PushKeyChar(#0);
               PushKeyChar(#81);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;               
               End;
             '1': Begin
               PushKeyChar(#0);
               PushKeyChar(#71);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;               
               End;
             '4': Begin
               PushKeyChar(#0);
               PushKeyChar(#79);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;               
               End;
             '3': Begin
               PushKeyChar(#0);
               PushKeyChar(#83);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;               
               End;
             '2': Begin
               PushKeyChar(#0);
               PushKeyChar(#82);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;
               End;
             Else
               Begin
               PushKeyChar(#27);
               PushKeyChar('[');
               PushKeyChar(TmpChar);
               State := 0;
               End;
             End;
           End;
        3: Begin { have esc [[ }
             Case TmpChar of
               'A': Begin
                 PushKeyChar(#0);
                 PushKeyChar(#59);
                 State := 0;
                 End;
               'B': Begin
                 PushKeyChar(#0);
                 PushKeyChar(#60);
                 State := 0;
                 End;
               'C': Begin
                 PushKeyChar(#0);
                 PushKeyChar(#61);
                 State := 0;
                 End;
               'D': Begin
                 PushKeyChar(#0);
                 PushKeyChar(#62);
                 State := 0;
                 End;
               'E': Begin
                 PushKeyChar(#0);
                 PushKeyChar(#63);
                 State := 0;
                 End;
               End;
           End;
        4: Begin { have esc [1 }
           Case TmpChar Of
             '7': Begin
               PushKeyChar(#0);
               PushKeyChar(#64);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;
               End;
             '8': Begin
               PushKeyChar(#0);
               PushKeyChar(#65);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;
               End;
             '9': Begin
               PushKeyChar(#0);
               PushKeyChar(#66);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;
               End;
             End;
           End;
        5: Begin { have esc [2 }
           Case TmpChar Of
             '0': Begin
               PushKeyChar(#0);
               PushKeyChar(#67);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;
               End;
             '1': Begin
               PushKeyChar(#0);
               PushKeyChar(#68);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;
               End;
             '3': Begin
               PushKeyChar(#0);
               PushKeyChar(#133);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;
               End;
             '4': Begin
               PushKeyChar(#0);
               PushKeyChar(#134);
               NumRead := Sys_Read(CrtFIn, @TmpChar, 1);
               State := 0;
               End;
             End;
          End;
        End;       
      End;
    End
  Else
    Begin
    If NumRead > 0 Then
      PushKeyChar(TmpChar);
    End;
  End;


Function  ReadKey: Char;
{
  Read key from internal buffer; If none available, get one from kernel.
}
  Begin
  If KeyPut <> KeySend Then
    ReadKey := PopKeyChar
  Else
    Begin
    ReadKeyDtl;
      ReadKey := PopKeyChar
    End;
  End;


Function ConvertString (InStr : String) : String;
{
  Convert a given string to a string which has escape codes, such that when
  output to the screen, the current window is respected.
  As little key-codes as possible are stored.
  When overflow threathens, part of the string is already dumped.
  DO NOT USE DIRECTLY !
}
var OutStr : string[255];
    i : Integer;
    
begin
Outstr:='';
if length (instr)>0 then
  for i:=1 to length(InStr) do
    begin
    inc(CurrX);
    if (CurrX>WinMaxX) or ((instr[i]=#10) or (instr[i]=#13)) then
      begin
      { When going to a new line, we dump the current one,
        this keeps the stringlength limited and less than 256...}
      if length(OutStr)>0 then
        DumpString(OutStr);
      CurrX:=WinMinX;
      if Instr[i]<>#13 then inc(CurrY);
      if CurrY>WinMaxY then
        begin
        CurrY:=WinMaxY;
        ScrollWindow(WinMinX,WinMinY,WinMaxX,WinMaxY,1);
        end;
      OutStr:=GenerateXY(CurrX,CurrY);
      dec(CurrX);
      end;
    if (InStr[i]<>#10) and (Instr[i]<>#13)  then
       begin
       OutStr:=OutStr+InStr[i];
       Scrn[CurrY,CurrX]:=Instr[i];
       ScrnCol[CurrY,CurrX]:=TextAttr;
       end;
    end;
ConvertString:=OutStr;
end;

Function CrtWrite(Var F: TextRec): Integer;
{
  Top level write function for CRT
}
  Var
    Temp : String[255];

Begin
  Move (F.BufPTR^[0],Temp[1],F.BufPos);
  temp[0]:=chr(F.BufPos);
  Dumpstring(Convertstring(temp));
  F.BufPos := 0;
  CrtWrite := 0; 
End;

Function CrtClose(Var F: TextRec): Integer;
{
  Close CRT associated file.
}
  Begin
  F.Mode := fmClosed;
  CrtClose := 0;
  End;

Function CrtOpen(Var F: TextRec): Integer;
{
  Open CRT associated file.
}
  Begin
  If F.Mode = fmOutput Then
    CrtOpen := 0
  Else
    CrtOpen := 5;
  End;


Function CrtRead(Var F: TextRec): Integer;
{
  Read from CRT associated file.
}
  Begin
  F.BufEnd := Sys_Read(LongInt(F.Handle), PChar(F.BufPtr), F.BufSize);
  F.BufPos := F.BufEnd;
  CrtWrite(F);
  CrtRead := 0;
  End;


Function CrtInOut(Var F: TextRec): Integer;
{
  InOut function for CRT associated file. 
}
  Begin
  Case F.Mode of
     fmInput: CrtInOut := CrtRead(F);
     fmOutput: CrtInOut := CrtWrite(F);
     End;
  End;


Procedure AssignCrt(Var F: Text);
{
  Assign a file to the console. All output on file goes to console instead.
}
  Begin
  TextRec(F).Mode := fmClosed; 
  TextRec(F).BufSize := SizeOf(TextBuf);
  TextRec(F).BufPtr := @TextRec(F).Buffer;
  TextRec(F).BufPos := 0;
  TextRec(F).OpenFunc := @CrtOpen;
  TextRec(F).InOutFunc := @CrtInOut;
  TextRec(F).FlushFunc := @CrtWrite;
  TextRec(F).CloseFunc := @CrtClose;
  TextRec(F).Name[0] := #0;
  End;


Procedure HighVideo;
{
  Set highlighted output.
}
  Begin
  TextAttr := TextAttr Or $08;
  DumpString(GenerateCol(TextAttr));
  End;


Procedure LowVideo;
{
  Set normal output
}
  Begin
  TextAttr := TextAttr And $f7;
  DumpString(GenerateCol(TextAttr));
  End;


Procedure NormVideo;
{
  Set normal back and foregroundcolors.
}
  Begin
  TextAttr := 07;
  DumpString(GenerateCol(TextAttr));
  End;


Procedure DelLine;
{
  Delete current line. Scroll subsequent lines up
}
  Begin
  ScrollScrnRegionUp(WinMinX, CurrY, WinMaxX, WinMaxY, 1);
  End;


Procedure InsLine;
{
  Insert line at current cursor position. Scroll subsequent lines down.
}
  Begin
  ScrollScrnRegionDown(WinMinX, CurrY, WinMaxX, WinMaxY, 1);
  End;


Procedure DoIbmPcChars(B: Boolean);
{
 Set IBM PC character attributes.
}
  Begin
  If b Then
    DumpString(#27'(U')
  Else
    DumpString(#27'(B');
  End;



Procedure SetRawMode(b: Boolean);
{
  Set RAW keyboard mode on/off.
  VERY dangerous !!
}
  Type IoRec = Record
    CiFlag: Word;
    CoFlag: Word;
    CcFlag: Word;
    ClFlag: Word;
    CLine: Char;
    ccc: Array[1..8] of Char;
    End;

  Var
    Termio: IoRec;
    Regs: SysCallRegs;

  Begin
  Regs.ebx := 0;
  Regs.ecx := $5405;
  Regs.edx := LongInt(@Termio);
  Syscall(syscall_nr_ioctl, regs);
  If b Then
    Termio.ClFlag := Termio.ClFlag and $fff5 {icanon 2 + echo 8}
  else
    Termio.ClFlag := Termio.ClFlag or $000a; {icanon 2 + echo 8}
  Regs.ecx := $5406;
  Regs.ebx := 0;
  Regs.edx := LongInt(@Termio);
  Syscall(syscall_nr_ioctl, Regs);
  End;


Procedure CrtExit;
{
  We need to restore normal keuyboard mode upon exit !!
}
  Begin
  ExitProc := ExitSave;
  SetRawMode(False);
  End;

Begin
FillChar(EmptyLine[1],80,' ');
EmptyLine[0]:=#80;
ExitSave := ExitProc;
ExitProc := @CrtExit;
SetRawMode(True);
AssignCrt(Output);
AssignCrt(Input);
TextRec(Output).Mode := fmOutput;
TextRec(Input).Mode := fmInput;
CrtFOut := TextRec(Output).Handle;
CrtFIn := TextRec(Input).Handle;
KeyPut := 1;
KeySend := 1;
CurrX := 1;
CurrY := 1;
ClrScr;
DumpString(GenerateCol(TextAttr));
End.

