Unit printer;

{  $Define PRINTERDEBUG}

Interface

Uses Linux,syscalls;

Const DefFile = '/tmp/PID.lst';

Var Lst : Text;

Procedure AssignLst ( Var F : text; ToFile : string[255]);
{
 Assigns to F a printing device. ToFile is a string with the following form:
 '|filename options'  : This sets up a pipe with the program filename,
                        with the given options
 'filename' : Prints to file filename. Filename can contain the string 'PID'
              (No Quotes), which will be replaced by the PID of your program.
              When closing lst, the file will be sent to lpr and deleted.
              (lpr should be in PATH)
                
 'filename|' Idem as previous, only the file is NOT sent to lpr, nor is it
             deleted.
             (useful for opening /dev/printer or for later printing)
 
 Lst is set up using '/tmp/PID.lst'. You can change this behaviour at 
 compile time, setting the DefFile constant.
}

Implementation

Const P_TOF   = 1; { Print to file }
      P_TOFNP = 2; { Print to File, don't spool }
      P_TOP   = 3; { Print to Pipe }

Var Lpr : String[255]; { Contains path to lpr binary, including terminating
                           null character }

    TheOldExitProc : pointer;
    
type       
    textbuf = array[0..127] of char;
    textrec = record
      handle : word;
      mode : word;
      bufsize : word;
      _private : word;
      bufpos : word;
      bufend : word;
      bufptr : ^textbuf;
      openfunc : pointer;
      inoutfunc : pointer;
      flushfunc : pointer;
      closefunc : pointer;
      userdata : array[1..16] of byte;
      name : string[255];
      buffer : textbuf;
      end;
                                                                                 
Procedure PrintAndDelete ( Var F : String[255]);

var i,j  : longint;
    p,pp : ppchar;
    
begin
  if lpr=#0 then exit;
  i:=sys_fork;
  if i<0 then 
     exit; { No printing was done. We leave the file where it is.}
  if i=0 then
     begin
     { We're in the child }
     getmem(p,8);
     if p=nil then 
        halt(127);
     pp:=p;
     pp^:=@lpr[1];
     pp:=pp+4;
     pp^:=@f[1];  
     sys_execve(@(lpr[1]),p,envp);
     { In trouble here ! }
     halt(128)
     end
  else
     begin
     { We're in the parent. }
     sys_waitpid (i,@j,0);
     if j<>0 then 
        exit;
     { Erase the file }
     sys_unlink(@(f[1]));
     end
end;

Procedure OpenLstPipe ( Var F : Text);

begin
  Popen (f,textrec(f).name,'W');
end;


Procedure OpenLstFile ( Var F : Text);

var i : integer;

begin
 {$ifdef PRINTERDEBUG}
 writeln ('Printer : In OpenLstFile');
 {$endif}
 If textrec(f).mode <> fmoutput then exit;
 textrec(f).userdata[15]:=0; { set Zero length flag }
 i:=sys_open (@(textrec(f).name[1]), (Open_WrOnly or Open_Creat), 438);
 if i<0 then
    textrec(f).mode:=fmclosed
  else
    textrec(f).handle:=i; 
end;

Procedure CloseLstFile ( Var F : Text);
begin
 {$ifdef PRINTERDEBUG}
 writeln ('Printer : In CloseLstFile');
 {$endif}
  sys_close (textrec(f).handle);
  { In case length is zero, don't print : lpr would give an error }
  if (textrec(f).userdata[15]=0) and (textrec(f).userdata[16]=P_TOF) then
     begin
     sys_unlink(@textrec(f).name[1]);
     exit
     end;
  { Non empty : needs printing ? } 
  if (textrec(f).userdata[16]=P_TOF) then 
     PrintAndDelete (textrec(f).name);
  textrec(f).mode:=fmclosed
end;

Procedure InOutLstFile ( Var F : text);

begin
 {$ifdef PRINTERDEBUG}
 writeln ('Printer : In InOutLstFile');
 {$endif}
 If textrec(f).mode<>fmoutput then exit;
 if textrec(f).bufpos<>0 then
    textrec(f).userdata[15]:=1; { To show that it is not empty. 
                                  Important when closing !!}
 sys_write (textrec(f).handle,pchar(textrec(f).bufptr),textrec(f).bufpos);
 textrec(f).bufpos:=0;
end;

Procedure SubstPidInName ( Var s : string);

var i : byte;
    temp : string[8];
    
begin
  i:=pos ('PID',s);
  if i=0 then exit;
  delete (s,i,3);
  str (sys_getpid,temp);
  insert (temp,s,i);
  {$ifdef PRINTERDEBUG}
  writeln ('Filename became : ',s);
  {$endif}
end;


Procedure AssignLst ( Var F : text; ToFile : string[255]);

begin
 {$ifdef PRINTERDEBUG}
 writeln ('Printer : In AssignLst');
 {$endif}
 If ToFile='' then 
    exit;
 textrec(f).bufptr:=@textrec(f).buffer;
 textrec(f).bufsize:=128;
 SubstPidInName (Tofile);
 if ToFile[1]='|' then
    begin
    textrec(f).name:=Tofile+#0;
    delete (Textrec(f).name,1,1);
    textrec(f).mode:=fmclosed;
    textrec(f).userdata[16]:=P_TOP;
    textrec(f).OpenFunc:=@OpenLstPipe;
    end
 else
    begin
    if Tofile[Length(ToFile)]='|' then
       begin
       textrec(f).userdata[16]:=P_TOFNP;
       textrec(f).name:=copy(ToFile,1,length(Tofile)-1)+#0
       end
    else
       begin
       textrec(f).userdata[16]:=P_TOF;
       textrec(f).name:=Tofile+#0
       end;
    textrec(f).OpenFunc:=@OpenLstFile;
    textrec(f).CloseFunc:=@CloseLstFile;
    textrec(f).InoutFunc:=@InoutLstFile;
    textrec(f).FlushFunc:=@InoutLstFile;
    end
end;

Procedure PrinterExitProc;

begin
  close(lst);
  ExitProc:=TheOldExitProc
end;

begin
  TheOldExitProc:=ExitProc;
  ExitProc:=@PrinterExitProc;
  AssignLst(Lst,DefFile);
  rewrite (Lst);
  lpr:='/usr/bin/lpr'#0
end.
                        